import math

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy.interpolate import interp1d

from .logs import get_logger, log_function_duration

logger = get_logger(__name__)


def calc_snr_per_row(s, N, exprs, exprs_sums, exprs_sq_sums):
    """Calculate SNR per row for given bicluster size.

    Args:
        s (int): bicluster size (number of samples)
        N (int): total number of samples
        exprs (array): expression matrix
        exprs_sums (array): precomputed row sums
        exprs_sq_sums (array): precomputed squared row sums

    Returns:
        array: SNR values per row
    """
    bic = exprs[:, :s]
    bic_sums = bic.sum(axis=1)
    bic_sq_sums = np.square(bic).sum(axis=1)

    bg_counts = N - s
    bg_sums = exprs_sums - bic_sums
    bg_sq_sums = exprs_sq_sums - bic_sq_sums

    bic_mean, bic_std = calc_mean_std_by_powers((s, bic_sums, bic_sq_sums))
    bg_mean, bg_std = calc_mean_std_by_powers((bg_counts, bg_sums, bg_sq_sums))

    snr_dist = (bic_mean - bg_mean) / (bic_std + bg_std)

    return snr_dist


def calc_mean_std_by_powers(powers):
    """Calculate mean and standard deviation from power statistics.

    Args:
        powers (tuple): tuple containing (count, sum, sum_of_squares)

    Returns:
        tuple: (mean, std) calculated from the power statistics
    """
    count, val_sum, sum_sq = powers

    mean = val_sum / count  # what if count == 0?
    std = np.sqrt((sum_sq / count) - mean * mean)
    return mean, std


def calc_SNR(ar1, ar2, pd_mode=False):
    """Calculate Signal-to-Noise Ratio (SNR) for two arrays.

    Args:
        ar1 (array): first array
        ar2 (array): second array
        pd_mode (bool): if True, use pandas-like mean/std methods
            i.e. n-1 for std, ignore nans

    Returns:
        float: SNR value
    """

    if pd_mode:
        std = lambda x: np.nanstd(x, ddof=1.0)
        mean = np.nanmean
    else:
        std = np.nanstd
        mean = np.mean

    mean_diff = mean(ar1) - mean(ar2)
    std_sum = std(ar1) + std(ar2)

    if std_sum == 0:
        return np.inf * mean_diff

    return mean_diff / std_sum


@log_function_duration(name="Background distribution generation")
def generate_null_dist(N, sizes, n_permutations=10000, pval=0.001, seed=42):
    """Generate null distribution of SNR values for statistical testing.

    Args:
        N (int): total number of samples
        sizes (array): bicluster sizes to generate null distributions for
        n_permutations (int): number of permutations for empirical distribution
        pval (float): p-value threshold for background distribution
        seed (int): random seed for reproducibility

    Returns:
        DataFrame: null distribution matrix with sizes as rows and permutations as columns
    """
    # samples 'N' values from standard normal distribution, and split them into bicluster and background groups
    # 'sizes' defines bicluster sizes to test
    # returns a dataframe with the distribution of SNR for each bicluster size (sizes x n_permutations )
    logger.debug(
        "Generate background distribution of SNR depending on the bicluster size:"
    )
    logger.debug(f"- total samples: {N}")
    logger.debug(f"- samples in a bicluster: {min(sizes)}-{max(sizes)}")
    logger.debug(f"- tn_permutations: {n_permutations}")
    logger.debug(f"- snr pval threshold: {pval}")

    exprs = np.zeros((n_permutations, N))  # generate random expressions from st.normal
    # values = exprs.values.reshape(-1) # random samples from expression matrix
    # exprs = np.random.choice(values,size=exprs.shape[1])
    np.random.seed(seed=seed)
    for i in range(n_permutations):
        exprs[i,] = sorted(np.random.normal(size=N))

    exprs_sums = exprs.sum(axis=1)
    exprs_sq_sums = np.square(exprs).sum(axis=1)

    null_distribution = pd.DataFrame(
        np.zeros((sizes.shape[0], n_permutations)),
        index=sizes,
        columns=range(n_permutations),
    )

    for s in sizes:
        null_distribution.loc[s, :] = -1 * calc_snr_per_row(
            s, N, exprs, exprs_sums, exprs_sq_sums
        )

    return null_distribution


def get_trend(sizes, thresholds, plot=True):
    """Smoothen the trend and return a function min_SNR(size; p-val cutoff).

    Given a set of points (x_i, y_i), returns a function f(x) that interpolates
    the data with LOWESS+linear interpolation.

    Args:
        sizes (array): values of x_i (bicluster sizes)
        thresholds (array): values of y_i (SNR thresholds)
        plot (bool): if True, plots the trend

    Returns:
        function: get_min_snr function that returns the minimal SNR for a given size
    """
    assert len(sizes) >= 0
    if len(sizes) == 1:
        return lambda x: thresholds[0]

    lowess = sm.nonparametric.lowess
    frac = max(1, min(math.floor(int(0.1 * len(sizes))), 15) / len(sizes))
    lowess_curve = lowess(
        thresholds, sizes, frac=frac, return_sorted=True, is_sorted=False
    )
    get_min_snr = interp1d(
        lowess_curve[:, 0], lowess_curve[:, 1]
    )  # ,kind="nearest-up",fill_value="extrapolate")
    if plot:
        plt.plot(sizes, thresholds, "b--", lw=2)
        plt.plot(sizes, get_min_snr(sizes), "r-", lw=2)
        plt.xlabel("n_samples")
        plt.ylabel("SNR")
        plt.ylim((0, 5))
        plt.show()
    return get_min_snr


def calc_e_pval(snr, size, null_distribution):
    """Calculate empirical p-value from null distribution.

    Args:
        snr (float): observed SNR value
        size (int): bicluster size
        null_distribution (DataFrame): precomputed null distribution

    Returns:
        float: empirical p-value
    """
    e_dist = null_distribution.loc[int(size), :]
    return (len(e_dist[e_dist >= abs(snr)]) + 1.0) / (null_distribution.shape[1] + 1.0)



"""Tests for binarization module."""

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from unpast.core import binarization
from unpast.utils.io import ProjectPaths, read_args


def test__select_pos_neg_gmm():
    # Two clear clusters
    row = np.array([1, 1, 1, 10, 10, 10])
    mask_pos, mask_neg, snr, size, is_converged = binarization._select_pos_neg(
        row, min_n_samples=2, method="GMM"
    )
    # Should split into two groups of size 3
    assert mask_pos.sum() == 3
    assert mask_neg.sum() == 3
    assert snr > 0
    assert size == 3


def test__select_pos_neg_kmeans():
    row = np.array([1, 1, 1, 10, 10, 10])
    mask_pos, mask_neg, snr, size, is_converged = binarization._select_pos_neg(
        row, min_n_samples=2, method="kmeans"
    )
    assert mask_pos.sum() == 3
    assert mask_neg.sum() == 3
    assert snr > 0
    assert size == 3


def test__select_pos_neg_too_few_samples():
    row = np.array([1, 1, 10])
    mask_pos, mask_neg, snr, size, is_converged = binarization._select_pos_neg(
        row, min_n_samples=4, method="GMM"
    )
    # Should not assign any group if not enough samples
    assert size != size or size < 4  # nan or < min_n_samples


def test__select_pos_neg_methods():
    for method in ["GMM", "kmeans", "ward"]:
        row = np.array([1, 1, 1, 10, 10, 10])
        mask_pos, mask_neg, snr, size, is_converged = binarization._select_pos_neg(
            row, min_n_samples=2, method=method
        )
        assert mask_pos.sum() == 3
        assert mask_neg.sum() == 3
        assert snr > 0
        assert size == 3

    with pytest.raises(RuntimeError):
        row = np.array([1, 1, 1, 10, 10, 10])
        binarization._select_pos_neg(row, min_n_samples=2, method="invalid_method")


def test_sklearn_binarization_basic():
    # 2 genes, 6 samples, clear separation
    data = pd.DataFrame(
        {
            "s1": [1, 10],
            "s2": [1, 10],
            "s3": [1, 10],
            "s4": [10, 1],
            "s5": [10, 1],
            "s6": [10, 1],
        },
        index=["geneA", "geneB"],
    )
    binarized, stats = binarization.sklearn_binarization(
        data, min_n_samples=2, plot=False
    )
    assert set(binarized.columns) == {"geneA", "geneB"}
    assert binarized.shape == (6, 2)
    assert all(x in stats.columns for x in ["SNR", "size", "direction"])


def test_sklearn_binarization_direction():
    # geneA: up-regulated, geneB: down-regulated
    data = pd.DataFrame(
        {
            "s1": [1, 10],
            "s2": [1, 10],
            "s3": [1, 10],
            "s4": [10, 1],
            "s5": [10, 1],
            "s6": [10, 1],
        },
        index=["geneA", "geneB"],
    )
    _, stats = binarization.sklearn_binarization(data, min_n_samples=2, plot=False)
    assert set(stats["direction"]) <= {"UP", "DOWN"}


def test_binarize_minimal(monkeypatch, tmp_path):
    # Patch plotting and file I/O
    monkeypatch.setattr(binarization, "plot_binarized_feature", lambda *a, **k: None)
    monkeypatch.setattr(binarization.pd.DataFrame, "to_csv", lambda *a, **k: None)
    # Create data with larger groups to ensure they pass the min_n_samples threshold
    data = pd.DataFrame(
        {
            "s1": [1, 10],
            "s2": [1, 10],
            "s3": [1, 10],
            "s4": [1, 10],
            "s5": [1, 10],
            "s6": [10, 1],
            "s7": [10, 1],
            "s8": [10, 1],
            "s9": [10, 1],
            "s10": [10, 1],
        },
        index=["geneA", "geneB"],
    )

    binarized, stats, null_dist = binarization.binarize(
        ProjectPaths(str(tmp_path / "testbin")),
        exprs=data,
        no_binary_save=True,
        plot_all=False,
        min_n_samples=3,
        n_permutations=100,
    )
    assert isinstance(binarized, pd.DataFrame)
    assert isinstance(stats, pd.DataFrame)
    assert null_dist is not None


def test_binarize_save_load_simple(tmp_path):
    """Test simple save and load functionality using internal save/load functions directly."""
    # Create simple test data
    binarized_data = pd.DataFrame(
        {
            "geneA": [1, 0, 1, 0],
            "geneB": [0, 1, 0, 1],
        },
        index=["s1", "s2", "s3", "s4"],
    )

    stats = pd.DataFrame(
        {
            "SNR": [2.5, 3.0],
            "size": [2, 2],
            "direction": ["UP", "DOWN"],
            "pval": [0.01, 0.005],
        },
        index=["geneA", "geneB"],
    )

    null_dist = pd.DataFrame(
        {
            2: [1.0, 1.2, 1.5],
            3: [1.1, 1.3, 1.6],
        },
        index=[0, 1, 2],
    )

    paths = ProjectPaths(str(tmp_path / "simple_test"))
    args_saveable = {
        "min_n_samples": 3,
        "paths": "1234",
    }

    # 1. Save - use internal save function
    binarization._save_binarization_files(
        paths, args_saveable, binarized_data, stats, null_dist
    )

    # Verify save worked by checking files exist
    assert Path(paths.binarization_res).exists()
    assert Path(paths.binarization_stats).exists()
    assert Path(paths.binarization_bg).exists()
    assert Path(paths.binarization_args).exists()

    # 2. Load - use internal load function to load saved files
    args_saveable["paths"] = "should be ignored"
    binarized2, stats2, null_dist2 = binarization._try_loading_binarization_files(
        paths, args_saveable
    )

    # Results should be identical when loaded from cache
    assert binarized2 is not None
    assert stats2 is not None
    assert null_dist2 is not None
    pd.testing.assert_frame_equal(binarized_data, binarized2, check_dtype=False)
    pd.testing.assert_frame_equal(stats, stats2, check_dtype=False)
    pd.testing.assert_frame_equal(null_dist, null_dist2, check_dtype=False)

    # 3. Load with changed args - should return None values (no cache hit)
    changed_args = args_saveable.copy()
    changed_args["min_n_samples"] = 4  # Different arg

    binarized3, stats3, null_dist3 = binarization._try_loading_binarization_files(
        paths, changed_args
    )

    # Should return None because args don't match
    assert binarized3 is None
    assert stats3 is None
    assert null_dist3 is None


def test_binarize_save_load_comprehensive(tmp_path):
    """Test comprehensive save and load functionality with file verification."""
    # Create test data with clear separation
    data = pd.DataFrame(
        {
            "s1": [1, 10, 1],
            "s2": [1, 10, 1],
            "s3": [1, 10, 1],
            "s4": [1, 10, 1],
            "s5": [1, 10, 1],
            "s6": [10, 1, 10],
            "s7": [10, 1, 10],
            "s8": [10, 1, 10],
            "s9": [10, 1, 10],
            "s10": [10, 1, 10],
        },
        index=["geneA", "geneB", "geneC"],
    )

    paths = ProjectPaths(str(tmp_path / "comprehensive_test"))

    # 1. Run binarization (first time - should save files)
    binarized1, stats1, null_dist1 = binarization.binarize(
        exprs=data,
        plot_all=False,
        paths=paths,
        min_n_samples=3,
        n_permutations=100,
        method="GMM",
        seed=42,
    )

    # Verify that files were created
    created_files = list(tmp_path.rglob("*.tsv"))
    assert len(created_files) >= 4, (
        f"Expected at least 4 files, got {len(created_files)}: {[f.name for f in created_files]}"
    )

    # Check specific files exist
    expected_files = [
        "bin_args.tsv",
        "bin_res.tsv",
        "bin_stats.tsv",
        "bin_background.tsv",
    ]
    actual_file_names = [f.name for f in created_files]
    for expected_file in expected_files:
        assert expected_file in actual_file_names, f"Missing file: {expected_file}"

    # 2. Try load files (should load from cached files)
    args_saveable = read_args(paths.binarization_args)
    binarized2, stats2, null_dist2 = binarization._try_loading_binarization_files(
        paths, args_saveable
    )

    # Results should be identical when loaded from cache
    assert binarized2 is not None
    assert stats2 is not None
    assert null_dist2 is not None

    pd.testing.assert_frame_equal(binarized1, binarized2, check_dtype=False)
    pd.testing.assert_frame_equal(stats1, stats2, check_dtype=False)
    pd.testing.assert_frame_equal(null_dist1, null_dist2, check_dtype=False)

    # 3. Try load files with changed args (should return None values)
    changed_args = args_saveable.copy()
    changed_args["min_n_samples"] = 4  # Changed argument
    changed_args["n_permutations"] = 50  # Changed argument
    changed_args["method"] = "kmeans"  # Changed argument

    binarized3, stats3, null_dist3 = binarization._try_loading_binarization_files(
        paths, changed_args
    )

    # Should return None because args don't match
    assert binarized3 is None
    assert stats3 is None
    assert null_dist3 is None


@pytest.mark.plot
def test_binarize_plot(monkeypatch, tmp_path):
    # Patch plotting and file I/O
    monkeypatch.setattr(binarization, "plot_binarized_feature", lambda *a, **k: None)
    monkeypatch.setattr(binarization.pd.DataFrame, "to_csv", lambda *a, **k: None)
    # Create data with larger groups to ensure they pass the min_n_samples threshold
    data = pd.DataFrame(
        {
            "s1": [1, 10],
            "s2": [1, 10],
            "s3": [1, 10],
            "s4": [1, 10],
            "s5": [1, 10],
            "s6": [10, 1],
            "s7": [10, 1],
            "s8": [10, 1],
            "s9": [10, 1],
            "s10": [10, 1],
        },
        index=["geneA", "geneB"],
    )

    binarized, stats, null_dist = binarization.binarize(
        ProjectPaths(str(tmp_path / "testbin")),
        exprs=data,
        no_binary_save=True,
        plot_all=True,
        min_n_samples=3,
        n_permutations=100,
        show_fits=["geneA", "geneB"],
        method="GMM",
        seed=42,
    )
    assert isinstance(binarized, pd.DataFrame)
    assert isinstance(stats, pd.DataFrame)
    assert null_dist is not None


if __name__ == "__main__":
    pytest.main([__file__])



"""Tests for feature_clustering module."""

import os
from unittest.mock import MagicMock, patch

import numpy as np
import pandas as pd
import pytest

from unpast.core.feature_clustering import run_Louvain
from unpast.utils.io import ProjectPaths
from unpast.utils.similarity import get_similarity_jaccard

TMP_DIR = os.path.join(os.path.dirname(__file__), "tmp_results")
if not os.access(TMP_DIR, os.W_OK):
    # repo dir is currently read-only during the testing stage in github-action
    TMP_DIR = "/tmp/unpast/tmp_results"


class TestRunLouvain:
    """Test cases for run_Louvain function."""

    def setup_method(self):
        """Set up test data."""
        # Create a similarity matrix with clear communities
        self.similarity_matrix = pd.DataFrame(
            {
                "gene1": [1.0, 0.8, 0.1, 0.1],
                "gene2": [0.8, 1.0, 0.1, 0.1],
                "gene3": [0.1, 0.1, 1.0, 0.9],
                "gene4": [0.1, 0.1, 0.9, 1.0],
            },
            index=["gene1", "gene2", "gene3", "gene4"],
        )

    def test_louvain_basic(self):
        """Test basic Louvain clustering."""
        modules, not_clustered, best_cutoff = run_Louvain(
            self.similarity_matrix,
            similarity_cutoffs=np.array([0.5]),
            plot=False,
        )

        # Should identify some modules
        assert isinstance(modules, list)
        assert isinstance(not_clustered, list)
        assert isinstance(best_cutoff, (float, int))

        # Total genes should be conserved
        total_genes = sum(len(module) for module in modules) + len(not_clustered)
        assert total_genes == 4

    def test_louvain_empty_similarity(self):
        """Test Louvain with empty similarity matrix."""
        empty_sim = pd.DataFrame({})

        modules, not_clustered, best_cutoff = run_Louvain(
            empty_sim, similarity_cutoffs=np.array([0.5]), plot=False
        )

        # Should handle empty input gracefully
        assert modules == []
        assert not_clustered == []
        assert best_cutoff is None

    def test_louvain_single_cutoff(self):
        """Test Louvain with single similarity cutoff."""
        modules, not_clustered, best_cutoff = run_Louvain(
            self.similarity_matrix,
            similarity_cutoffs=np.array([0.7]),
            plot=True,
        )

        # Should use the single cutoff provided
        assert best_cutoff == 0.7

    def test_louvain_multiple_cutoffs(self):
        """Test Louvain with multiple similarity cutoffs."""
        modules, not_clustered, best_cutoff = run_Louvain(
            self.similarity_matrix,
            similarity_cutoffs=np.array([0.3, 0.5, 0.7]),
            plot=False,
        )

        # Should select optimal cutoff automatically
        assert best_cutoff in [0.3, 0.5, 0.7]


class TestWGCNAFunctions:
    """Test cases for WGCNA-related functions."""

    def test_run_wgcna_parameter_validation(self, tmp_path):
        """Test WGCNA parameter validation without actually running R."""
        from unpast.core.feature_clustering import run_WGCNA

        # Create mock data
        data = pd.DataFrame(
            {
                "gene1": [1, 0, 1, 0],
                "gene2": [0, 1, 0, 1],
            },
            index=["1", "2", "3", "4"],
        )

        # Test invalid deepSplit parameter
        modules, not_clustered = run_WGCNA(
            data,
            paths=ProjectPaths(str(tmp_path)),
            deepSplit=5,  # Invalid value
        )

        # Should return empty results for invalid parameters
        assert modules == []
        assert not_clustered == []

        # Test invalid detectCutHeight parameter
        modules, not_clustered = run_WGCNA(
            data,
            paths=ProjectPaths(str(tmp_path)),
            detectCutHeight=1.5,  # Invalid value
        )

        # Should return empty results for invalid parameters
        assert modules == []
        assert not_clustered == []

    @patch("os.remove")
    @patch("pandas.read_csv")
    @patch("subprocess.Popen")
    def test_run_wgcna_mocked(self, mock_popen, mock_read_csv, mock_remove, tmp_path):
        """Test WGCNA execution with mocked subprocess and file operations."""
        from unpast.core.feature_clustering import run_WGCNA

        # Mock the subprocess call
        mock_process = MagicMock()
        mock_process.communicate.return_value = (b"", b"")
        mock_popen.return_value = mock_process

        # Mock the CSV reading to return a sample modules result
        mock_modules_df = pd.DataFrame(
            {"genes": ["gene1 gene2", "gene3"]}, index=[1, 2]
        )
        mock_read_csv.return_value = mock_modules_df

        # Create test data
        data = pd.DataFrame(
            {
                "gene1": [1, 0, 1, 0],
                "gene2": [0, 1, 0, 1],
                "gene3": [1, 1, 0, 0],
            },
            index=["1", "2", "3", "4"],
        )

        modules, not_clustered = run_WGCNA(
            data,
            paths=ProjectPaths(str(tmp_path)),
            deepSplit=2,
            detectCutHeight=0.8,
        )

        # Verify subprocess was called
        assert mock_popen.called

        # Should return parsed modules
        assert len(modules) >= 0  # Depends on mock data
        assert isinstance(not_clustered, list)

    @patch("subprocess.Popen")
    @patch("pandas.read_csv")
    def test_run_wgcna_file_error(self, mock_read_csv, mock_popen, tmp_path):
        """Test WGCNA when output file cannot be read."""
        from unpast.core.feature_clustering import run_WGCNA
        from unpast.utils.io import ProjectPaths

        # Mock subprocess
        mock_process = MagicMock()
        mock_process.communicate.return_value = (b"", b"some error")
        mock_popen.return_value = mock_process

        # Mock CSV reading to raise an exception
        mock_read_csv.side_effect = Exception("File not found")

        data = pd.DataFrame(
            {
                "gene1": [1, 0, 1, 0],
                "gene2": [0, 1, 0, 1],
            },
            index=["1", "2", "3", "4"],
        )
        paths = ProjectPaths(str(tmp_path))
        modules, not_clustered = run_WGCNA(
            data,
            paths=paths,
        )

        # Should handle file read errors gracefully
        assert modules == []
        assert not_clustered == []

    def test_run_wgcna_iterative_basic(self, tmp_path):
        """Test basic functionality of run_WGCNA_iterative."""
        from unpast.core.feature_clustering import run_WGCNA_iterative

        # Create test data
        data = pd.DataFrame(
            {
                "gene1": [1, 0, 1, 0],
                "gene2": [0, 1, 0, 1],
                "gene3": [1, 1, 0, 0],
            },
            index=["1", "2", "3", "4"],
        )
        from unpast.utils.io import ProjectPaths

        # Mock run_WGCNA to avoid R dependencies
        with patch("unpast.core.feature_clustering.run_WGCNA") as mock_wgcna:
            # Mock to return no modules (stop condition)
            mock_wgcna.return_value = ([], ["gene1", "gene2", "gene3"])

            paths = ProjectPaths(str(tmp_path))
            modules, not_clustered = run_WGCNA_iterative(data, paths=paths)

            # Should call run_WGCNA at least once
            assert mock_wgcna.called
            assert isinstance(modules, list)
            assert isinstance(not_clustered, (list, np.ndarray))

    @pytest.mark.parametrize("method", ["WGCNA", "IterativeWGCNA"])
    def test_feature_name_handling(self, method, tmp_path):
        """Test handling of special characters in feature names."""
        from unpast.core.feature_clustering import run_WGCNA, run_WGCNA_iterative

        # Create data with spaces and duplicates in feature names
        data = pd.DataFrame(
            {
                "gene 1": [1, 0, 1, 0],
                "gene 2": [0, 1, 0, 1],
                "gene1": [1, 1, 0, 0],  # Duplicate after space removal
            },
            index=["v1", "v 2", "v 3", "v 3"],  # Duplicate feature names
        )

        # Mock subprocess and file operations to avoid R dependencies
        with (
            patch("subprocess.Popen") as mock_popen,
            patch("pandas.read_csv") as mock_read_csv,
            patch("os.remove"),
        ):
            mock_process = MagicMock()
            mock_process.communicate.return_value = (b"", b"")
            mock_popen.return_value = mock_process

            # Mock empty modules result
            mock_read_csv.side_effect = Exception("File not found")

            wgcna_func = {
                "WGCNA": run_WGCNA,
                "IterativeWGCNA": run_WGCNA_iterative,
            }[method]

            # Should handle name/index processing without crashing
            for d in data, data.T:
                modules, not_clustered = wgcna_func(
                    d,
                    paths=ProjectPaths(str(tmp_path)),
                )
                assert isinstance(modules, list)
                assert isinstance(not_clustered, list)


class TestIntegrationFeatureClustering:
    """Integration tests for feature clustering pipeline."""

    def test_similarity_to_clustering_pipeline(self):
        """Test the pipeline from similarity calculation to clustering."""
        # Create binary data with clear patterns
        data = pd.DataFrame(
            {
                "gene1": [1, 1, 0, 0, 0, 0],
                "gene2": [1, 1, 1, 0, 0, 0],  # Similar to gene1
                "gene3": [0, 0, 0, 1, 1, 1],  # Different cluster
                "gene4": [0, 0, 0, 1, 1, 0],  # Similar to gene3
            },
            index=["1", "2", "3", "4", "5", "6"],
        )

        # Calculate similarity
        similarity = get_similarity_jaccard(data)

        # Run clustering
        modules, not_clustered, best_cutoff = run_Louvain(
            similarity,
            similarity_cutoffs=np.array([0.3, 0.5]),
            plot=False,
        )

        # Should identify some structure
        total_genes = sum(len(module) for module in modules) + len(not_clustered)
        assert total_genes == 4

        # If clustering worked, should have at least one module
        if len(modules) > 0:
            assert all(len(module) >= 1 for module in modules)


if __name__ == "__main__":
    pytest.main([__file__])



"""Tests for sample_clustering module."""

from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from unpast.core.sample_clustering import (
    cluster_samples,
    make_biclusters,
    merge_biclusters,
    modules2biclusters,
    update_bicluster_data,
)


class TestClusterSamples:
    """Test cases for cluster_samples function."""

    def test_cluster_samples_basic(self):
        """Test basic clustering functionality."""
        # Create data with clear separation
        # 4 samples with high values, 4 samples with low values
        data = pd.DataFrame(
            {"gene1": [1, 1, 1, 1, 10, 10, 10, 10], "gene2": [2, 2, 2, 2, 8, 8, 8, 8]},
            index=["s1", "s2", "s3", "s4", "s5", "s6", "s7", "s8"],
        )

        result = cluster_samples(data, min_n_samples=3, method="kmeans")

        # Should identify the smaller cluster (4 samples)
        assert len(result) > 0
        assert result["n_samples"] == 4
        assert len(result["sample_indexes"]) == 4

    def test_cluster_samples_methods(self):
        """Test different clustering methods."""
        data = pd.DataFrame(
            {"gene1": [1, 1, 1, 10, 10, 10], "gene2": [2, 2, 2, 8, 8, 8]},
            index=["s1", "s2", "s3", "s4", "s5", "s6"],
        )

        methods = ["kmeans", "ward", "GMM"]
        for method in methods:
            result = cluster_samples(data, min_n_samples=2, method=method)
            assert len(result) > 0
            assert result["n_samples"] == 3
            assert len(result["sample_indexes"]) == 3

    def test_cluster_samples_insufficient_samples(self):
        """Test when there are insufficient samples."""
        data = pd.DataFrame(
            {"gene1": [1, 1, 10], "gene2": [2, 2, 8]}, index=["s1", "s2", "s3"]
        )

        result = cluster_samples(data, min_n_samples=5, method="kmeans")

        # Should return empty dict if not enough samples
        assert len(result) == 0

    def test_cluster_samples_equal_groups(self):
        """Test when both groups have equal size."""
        data = pd.DataFrame(
            {"gene1": [1, 1, 10, 10], "gene2": [2, 2, 8, 8]},
            index=["s1", "s2", "s3", "s4"],
        )

        result = cluster_samples(data, min_n_samples=2, method="kmeans")

        # Should identify one of the groups
        assert len(result) > 0
        assert result["n_samples"] == 2
        assert len(result["sample_indexes"]) == 2


class TestModules2Biclusters:
    """Test cases for modules2biclusters function."""

    def setup_method(self):
        """Set up test data."""
        # Create expression data
        self.data = pd.DataFrame(
            {
                "s1": [1, 1, 5, 5],
                "s2": [1, 1, 5, 5],
                "s3": [1, 1, 5, 5],
                "s4": [10, 10, 1, 1],
                "s5": [10, 10, 1, 1],
                "s6": [10, 10, 1, 1],
            },
            index=["gene1", "gene2", "gene3", "gene4"],
        )

        # Create modules (gene clusters)
        self.modules = [
            ["gene1", "gene2"],  # First module
            ["gene3", "gene4"],  # Second module
            ["gene1"],  # Single gene module (should be filtered)
        ]

    def test_modules2biclusters_basic(self):
        """Test basic module to bicluster conversion."""
        biclusters = modules2biclusters(
            self.modules, self.data, min_n_samples=2, min_n_genes=2
        )

        # Should create 2 biclusters (single gene module filtered out)
        assert len(biclusters) == 2

        # Check structure of biclusters
        for i, bic in biclusters.items():
            assert "id" in bic
            assert "genes" in bic
            assert "n_genes" in bic
            assert "sample_indexes" in bic
            assert "n_samples" in bic
            assert bic["n_genes"] >= 2
            assert bic["n_samples"] >= 2

    def test_modules2biclusters_min_genes_filter(self):
        """Test minimum genes filtering."""
        biclusters = modules2biclusters(
            self.modules,
            self.data,
            min_n_samples=2,
            min_n_genes=3,  # Higher threshold
        )

        # Should create 0 biclusters (all modules have < 3 genes)
        assert len(biclusters) == 0

    def test_modules2biclusters_min_samples_filter(self):
        """Test minimum samples filtering."""
        biclusters = modules2biclusters(
            self.modules,
            self.data,
            min_n_samples=10,  # Higher than available samples
            min_n_genes=2,
        )

        # Should create 0 biclusters (not enough samples)
        assert len(biclusters) == 0


class TestUpdateBiclusterData:
    """Test cases for update_bicluster_data function."""

    def setup_method(self):
        """Set up test data."""
        # Create expression data with clear up/down regulation
        self.data = pd.DataFrame(
            {
                "s1": [2, -2, 0],  # High gene1, low gene2
                "s2": [2, -2, 0],  # High gene1, low gene2
                "s3": [0, 0, 0],  # Background
                "s4": [0, 0, 0],  # Background
            },
            index=["gene1", "gene2", "gene3"],
        )

        # Create a basic bicluster
        self.bicluster = {
            "sample_indexes": {0, 1},  # s1, s2
            "genes": {"gene1", "gene2"},
            "n_genes": 2,
            "n_samples": 2,
        }

    def test_update_bicluster_data_basic(self):
        """Test basic bicluster data update."""
        result = update_bicluster_data(self.bicluster, self.data)

        # Check new fields are added
        assert "samples" in result
        assert "gene_indexes" in result
        assert "genes_up" in result
        assert "genes_down" in result
        assert "SNR" in result

        # Check sample names
        assert result["samples"] == {"s1", "s2"}

        # Check gene regulation
        assert "gene1" in result["genes_up"]  # Should be up-regulated
        assert "gene2" in result["genes_down"]  # Should be down-regulated

    def test_update_bicluster_data_snr(self):
        """Test SNR calculation."""
        result = update_bicluster_data(self.bicluster, self.data)

        # SNR should be calculated and positive
        assert result["SNR"] > 0
        assert not np.isnan(result["SNR"])


class TestMakeBiclusters:
    """Test cases for make_biclusters function."""

    def setup_method(self):
        """Set up test data."""
        # Create expression data
        self.data = pd.DataFrame(
            {
                "s1": [2, 2, 0],
                "s2": [2, 2, 0],
                "s3": [0, 0, 2],
                "s4": [0, 0, 2],
            },
            index=["gene1", "gene2", "gene3"],
        )

        # Create binarized data
        self.binarized_data = pd.DataFrame(
            {"gene1": [1, 1, 0, 0], "gene2": [1, 1, 0, 0], "gene3": [0, 0, 1, 1]},
            index=["s1", "s2", "s3", "s4"],
        )

        # Create feature clusters
        self.feature_clusters = [["gene1", "gene2"], ["gene3"]]

    def test_make_biclusters_basic(self):
        """Test basic bicluster creation."""
        biclusters = make_biclusters(
            self.feature_clusters,
            self.binarized_data,
            self.data,
            merge=1,  # No merging
            min_n_samples=2,
            min_n_genes=1,
        )

        # Should create a DataFrame with biclusters
        assert isinstance(biclusters, pd.DataFrame)
        assert len(biclusters) > 0

        # Check required columns
        required_cols = ["SNR", "n_genes", "n_samples", "genes", "samples", "direction"]
        for col in required_cols:
            assert col in biclusters.columns

    def test_make_biclusters_empty_input(self):
        """Test with empty feature clusters."""
        # This test reveals a bug in the original code when handling empty biclusters
        # The code tries to access columns that don't exist in an empty DataFrame
        with pytest.raises(KeyError):
            make_biclusters(
                [],  # Empty feature clusters
                self.binarized_data,
                self.data,
                merge=1,
                min_n_samples=2,
                min_n_genes=1,
            )

    def test_make_biclusters_direction(self):
        """Test direction assignment."""
        biclusters = make_biclusters(
            self.feature_clusters,
            self.binarized_data,
            self.data,
            merge=1,
            min_n_samples=2,
            min_n_genes=1,
        )

        # Check that direction is assigned
        if len(biclusters) > 0:
            directions = set(biclusters["direction"])
            assert directions.issubset({"UP", "DOWN", "BOTH"})

    def test_make_biclusters_sorting(self):
        """Test that biclusters are sorted by SNR and n_genes."""
        # Create data that will generate multiple biclusters
        data = pd.DataFrame(
            {
                "s1": [3, 1, 0],
                "s2": [3, 1, 0],
                "s3": [0, 2, 1],
                "s4": [0, 2, 1],
            },
            index=["gene1", "gene2", "gene3"],
        )

        binarized_data = pd.DataFrame(
            {"gene1": [1, 1, 0, 0], "gene2": [0, 0, 1, 1], "gene3": [0, 0, 1, 1]},
            index=["s1", "s2", "s3", "s4"],
        )

        feature_clusters = [["gene1"], ["gene2", "gene3"]]

        biclusters = make_biclusters(
            feature_clusters,
            binarized_data,
            data,
            merge=1,
            min_n_samples=2,
            min_n_genes=1,
        )

        if len(biclusters) > 1:
            # Check that biclusters are sorted by SNR (descending)
            snr_values = biclusters["SNR"].values
            assert all(
                snr_values[i] >= snr_values[i + 1] for i in range(len(snr_values) - 1)
            )


class TestMergeBiclusters:
    """Test cases for merge_biclusters function."""

    def setup_method(self):
        """Set up test data."""
        # Create expression data
        self.data = pd.DataFrame(
            {
                "s1": [1, 1],
                "s2": [1, 1],
                "s3": [1, 1],
                "s4": [10, 10],
                "s5": [10, 10],
                "s6": [10, 10],
            },
            index=["gene1", "gene2"],
        )

        # Create biclusters with overlapping samples
        self.biclusters = {
            0: {
                "sample_indexes": {0, 1, 2},  # s1, s2, s3
                "genes": {"gene1"},
                "n_genes": 1,
                "n_samples": 3,
            },
            1: {
                "sample_indexes": {0, 1},  # s1, s2 (overlapping)
                "genes": {"gene2"},
                "n_genes": 1,
                "n_samples": 2,
            },
            2: {
                "sample_indexes": {3, 4, 5},  # s4, s5, s6 (different)
                "genes": {"gene1"},
                "n_genes": 1,
                "n_samples": 3,
            },
        }

    def test_merge_biclusters_high_threshold(self):
        """Test merging with high Jaccard threshold (no merging expected)."""
        # Mock the required functions that are imported from feature_clustering
        mock_similarity = pd.DataFrame(
            {0: [1.0, 0.2, 0.1], 1: [0.2, 1.0, 0.1], 2: [0.1, 0.1, 1.0]},
            index=[0, 1, 2],
        )

        mock_merged = []  # No merging expected
        mock_not_merged = [0, 1, 2]  # All remain separate

        with (
            patch(
                "unpast.core.sample_clustering.get_similarity_jaccard"
            ) as mock_jaccard,
            patch("unpast.core.sample_clustering.run_Louvain") as mock_louvain,
        ):
            mock_jaccard.return_value = mock_similarity
            mock_louvain.return_value = (mock_merged, mock_not_merged, 0.9)

            result = merge_biclusters(
                self.biclusters,
                self.data,
                J=0.9,  # High threshold
                min_n_samples=2,
            )

            # Should return similar number of biclusters (no merging)
            assert len(result) == len(self.biclusters)
            assert isinstance(result, dict)

            # Check that the functions were called
            assert mock_jaccard.called
            assert mock_louvain.called

    def test_merge_biclusters_low_threshold(self):
        """Test merging with low Jaccard threshold."""
        # Mock the required functions
        mock_similarity = pd.DataFrame(
            {0: [1.0, 0.8, 0.1], 1: [0.8, 1.0, 0.1], 2: [0.1, 0.1, 1.0]},
            index=[0, 1, 2],
        )

        mock_merged = [[0, 1]]  # Biclusters 0 and 1 should be merged
        mock_not_merged = [2]  # Bicluster 2 remains separate

        with (
            patch(
                "unpast.core.sample_clustering.get_similarity_jaccard"
            ) as mock_jaccard,
            patch("unpast.core.sample_clustering.run_Louvain") as mock_louvain,
            patch("unpast.core.sample_clustering.cluster_samples") as mock_cluster,
        ):
            mock_jaccard.return_value = mock_similarity
            mock_louvain.return_value = (mock_merged, mock_not_merged, 0.1)

            # Mock cluster_samples to return a valid result for merged bicluster
            mock_cluster.return_value = {"sample_indexes": {0, 1, 2}, "n_samples": 3}

            result = merge_biclusters(
                self.biclusters,
                self.data,
                J=0.1,  # Low threshold
                min_n_samples=2,
            )

            # Should potentially merge some biclusters
            assert isinstance(result, dict)
            assert len(result) <= len(self.biclusters)

            # Check that the functions were called
            assert mock_jaccard.called
            assert mock_louvain.called


if __name__ == "__main__":
    pytest.main([__file__])



import warnings

import numpy as np
import pytest

from unpast.utils.statistics import calc_SNR, get_trend


def test_get_trend_single_point():
    sizes = [10]
    thresholds = [2.5]
    min_snr = get_trend(sizes, thresholds, plot=False)
    assert min_snr(10) == 2.5


@pytest.mark.plot
def test_get_trend_multiple_points():
    sizes = [10, 20, 30, 40, 50]
    thresholds = [2.5, 3.0, 3.5, 4.0, 4.5]
    min_snr = get_trend(sizes, thresholds, plot=True)
    assert np.allclose(min_snr(sizes), thresholds)


def test_get_trend_noisy():
    sizes = [1] * 100 + [2] * 100
    thresholds = np.linspace(0, 1, 200)
    min_snr = get_trend(sizes, thresholds, plot=False)
    # 0.1 tolerance, as set of used points may be not very big
    assert np.allclose(min_snr([1, 1.5, 2]), [0.25, 0.5, 0.75], atol=0.1)


def test_calc_SNR():
    val1 = calc_SNR([0, 1, 2], [1, 2, 3])
    val2 = calc_SNR([0, 1, 2], [1, 2, 3], pd_mode=True)
    np.testing.assert_almost_equal(val1, -0.6123724356957945)
    np.testing.assert_almost_equal(val2, -0.5)

    # assert not failing
    assert calc_SNR([0, 0, 0], [1, 1, 1]) == float("-inf")
    assert calc_SNR([0, 0, 0], [1, 1, 1], True) == float("-inf")
    assert calc_SNR([1, 1, 1], [0, 0, 0]) == float("+inf")

    # sends warning "overflow encountered"
    # big numbers
    big_nums = [1e307, 1e307, 1e307]
    small_nums = [1e-150, 0, 0]
    assert np.std(big_nums) < float("+inf")
    assert np.std(small_nums) > 0
    assert calc_SNR(big_nums, small_nums) in [float("inf"), float("+inf")]

    with warnings.catch_warnings():
        warnings.simplefilter("error")

        # zero std
        assert calc_SNR([1, 1, 1], [0, 0, 0]) in [float("inf"), float("+inf")]
        assert calc_SNR([0, 0, 0], [1, 1, 1]) in [float("inf"), float("-inf")]


# def test_zscore():
#     # Test case 1: Basic functionality
#     df = pd.DataFrame({
#         'A': [1, 2, 3],
#         'B': [4, 5, 6],
#         'C': [7, 8, 9]
#     })
#     result = zscore(df)
#     assert result.shape == df.shape
#     assert np.allclose(result.mean(), 0, atol=1e-7)
#     assert np.allclose(result.std(), 1, atol=1e-7)

#     # Test case 2: Handling zero variance
#     df_zero_var = pd.DataFrame({
#         'A': [1, 1, 1],
#         'B': [2, 3, 4],
#         'C': [5, 6, 7]
#     })
#     result_zero_var = zscore(df_zero_var)
#     assert result_zero_var.shape == df_zero_var.shape
#     assert np.allclose(result_zero_var.loc['A'], 0)

# def test_prepare_input_matrix():
#     # Test case 1: Basic functionality
#     df = pd.DataFrame({
#         'A': [1, 2, 3, 4, 5],
#         'B': [2, 3, 4, 5, 6],
#         'C': [3, 4, 5, 6, 7]
#     })
#     result = prepare_input_matrix(df)
#     assert result.shape == df.shape
#     assert np.allclose(result.mean(), 0, atol=1e-7)
#     assert np.allclose(result.std(), 1, atol=1e-7)

#     # Test case 2: Handling zero variance
#     df_zero_var = pd.DataFrame({
#         'A': [1, 1, 1, 1, 1],
#         'B': [2, 3, 4, 5, 6],
#         'C': [3, 4, 5, 6, 7]
#     })
#     result_zero_var = prepare_input_matrix(df_zero_var)
#     assert result_zero_var.shape == (2, 5)  # One row should be dropped

#     # Test case 3: Handling missing values
#     df_missing = pd.DataFrame({
#         'A': [1, 2, np.nan, 4, 5],
#         'B': [2, 3, 4, np.nan, 6],
#         'C': [3, 4, 5, 6, 7]
#     })
#     result_missing = prepare_input_matrix(df_missing, min_n_samples=4)
#     assert result_missing.shape == (2, 5)  # One row should be dropped

#     # Test case 4: Ceiling
#     df_ceiling = pd.DataFrame({
#         'A': [1, 2, 3, 4, 5],
#         'B': [2, 3, 4, 5, 6],
#         'C': [3, 4, 5, 6, 7]
#     })
#     result_ceiling = prepare_input_matrix(df_ceiling, ceiling=2)
#     assert result_ceiling.max().max() <= 2
#     assert result_ceiling.min().min() >= -2

#     # Test case 5: Non-standardized input
#     df_non_std = pd.DataFrame({
#         'A': [10, 20, 30, 40, 50],
#         'B': [20, 30, 40, 50, 60],
#         'C': [30, 40, 50, 60, 70]
#     })
#     result_non_std = prepare_input_matrix(df_non_std)
#     assert np.allclose(result_non_std.mean(), 0, atol=1e-7)
#     assert np.allclose(result_non_std.std(), 1, atol=1e-7)



import os
import time

from unpast.utils import logs


def test_log_file_creation_and_write(tmp_path):
    log_path = tmp_path / "test.log"
    logs.setup_logging(log_file=str(log_path))
    logger = logs.get_logger(__name__)
    logger.debug("Test log message")
    assert os.path.exists(log_path)
    with open(log_path) as f:
        content = f.read()
    assert "Test log message" in content


def test_log_file_append(tmp_path):
    log_path = tmp_path / "test_append.log"
    # First write using file append
    with open(log_path, "w") as f:
        f.write("First message\n")

    logs.setup_logging(log_file=str(log_path))
    logger = logs.get_logger(__name__)
    logger.debug("Second message")

    with open(log_path) as f:
        lines = f.read()

    assert "First message" in lines
    assert "Second message" in lines


def test_log_file_custom_level(tmp_path):
    log_path = tmp_path / "test_level.log"
    logs.setup_logging(log_file=str(log_path), log_file_level=40)  # 40 = ERROR
    logger = logs.get_logger(__name__)
    logger.error("Error message")
    with open(log_path) as f:
        content = f.read()
    assert "Error message" in content
    logger.debug("Info message")
    with open(log_path) as f:
        content = f.read()
    assert "Info message" not in content


def test_log_function_duration(tmp_path):
    log_path = tmp_path / "test_duration.log"
    logs.setup_logging(log_file=str(log_path))

    @logs.log_function_duration(name="timed_func")
    def timed_func():
        time.sleep(0.1)
        return 42

    result = timed_func()
    assert result == 42
    with open(log_path) as f:
        content = f.read()
    assert "timed_func completed in:" in content



"""Cluster binarized genes"""

import sys
from time import time

import numpy as np
import pandas as pd

from unpast.utils.logs import get_logger, log_function_duration

logger = get_logger(__name__)


@log_function_duration(name="Jaccard Similarity")
def get_similarity_jaccard(binarized_data):  # ,J=0.5
    """Calculate Jaccard similarity matrix between features based on binary expression patterns.

    Args:
        binarized_data (DataFrame): binary expression matrix with samples as rows and features as columns

    Returns:
        DataFrame: symmetric similarity matrix with Jaccard coefficients between all feature pairs
    """
    genes = binarized_data.columns.values
    n_samples = binarized_data.shape[0]
    size_threshold = int(min(0.45 * n_samples, (n_samples) / 2 - 10))
    # print("size threshold",size_threshold)
    n_genes = binarized_data.shape[1]
    df = np.array(binarized_data.T, dtype=bool)
    results = np.zeros((n_genes, n_genes))
    for i in range(0, n_genes):
        results[i, i] = 1
        g1 = df[i]

        for j in range(i + 1, n_genes):
            g2 = df[j]
            o = g1 * g2
            u = g1 + g2
            jaccard = o.sum() / u.sum()
            # try matching complements
            if g1.sum() > size_threshold:
                g1_complement = ~g1
                o = g1_complement * g2
                u = g1_complement + g2
                jaccard_c = o.sum() / u.sum()
            elif g2.sum() > size_threshold:
                g2 = ~g2
                o = g1 * g2
                u = g1 + g2
                jaccard_c = o.sum() / u.sum()
            else:
                jaccard_c = 0
            jaccard = max(jaccard, jaccard_c)
            results[i, j] = jaccard
            results[j, i] = jaccard

    results = pd.DataFrame(data=results, index=genes, columns=genes)
    logger.debug(
        f"Jaccard similarities for {binarized_data.shape[1]} features computed."
    )
    return results


# @log_function_duration(name="Pearson Similarity")

# def get_similarity_corr(df, verbose=True):
#     """Calculate correlation-based similarity matrix between features.

#     Args:
#         df (DataFrame): expression matrix with features as columns
#         verbose (bool): whether to print progress information

#     Returns:
#         DataFrame: correlation similarity matrix with positive correlations only
#     """
#     corr = df.corr()  # .applymap(abs)
#     corr = corr[corr > 0]  # to consider only direct correlations
#     corr = corr.fillna(0)
#     return corr



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.pyplot import gcf


def draw_heatmap2(
    exprs,
    biclusters=pd.DataFrame(),
    annot=None,
    color_dict=None,
    figsize=(20, 10),
    dendrogram_ratio=(0.01, 0.2),  # space for dendrogram
    colors_ratio=(0.005, 0.02),
    bicluster_colors="black",  #
    no_bic_columns=False,  # do not show bicluster annotation in columns
    legend_n_cols=0,  # how many columns in legend; default - 0, no legend
    no_cbar=False,
    cluster_rows=True,
    cluster_cols=False,  # enable hierarchical clustering of columns
    cluster_columns=True,
    xlabel="samples",
    col_labels=True,
    row_labels=False,
    color_range=(-3, 3),
    bic_prefix="bic_",
    plot_bg_genes=False,
    no_row_colors=True,
    highlight_row_labels=[],
    row_labels_black=False,
):
    """* exprs - expressions of genes to plot
    * biclusters in UnPaSt format
    * annot - annotation for samples, columns - subtypes, rows - samples
    * color_dict - how to color each label
    * bicluster_colors - color for bicluster annotation "black","auto" or list
    """
    bic_names = []
    ordered_genes = []
    row_colors = None
    col_colors = None
    sample_order = exprs.columns.values
    if type(annot) != type(None):
        cols = list(annot.columns.values)
        if not no_row_colors:
            row_colors = pd.DataFrame(
                data="white",
                index=exprs.index.values,
                columns=[bic_prefix + str(x) for x in biclusters.index.values],
            )

    else:
        annot = pd.DataFrame(index=exprs.columns)
        if not no_row_colors:
            row_colors = pd.DataFrame(index=exprs.index.values)
        cols = []
    if type(color_dict) != type(None):
        pass

    # list of bicluster colors
    if biclusters.shape[0] > 0:
        bic_colors = []
        if bicluster_colors == "black" or bicluster_colors == "redblue":
            bic_colors = ["black"] * biclusters.shape[0]
        elif bicluster_colors == "auto":
            palette = sns.color_palette("colorblind")
            # Get the first n colors from the palette
            bic_colors = sns.color_palette("colorblind", biclusters.shape[0]).as_hex()
            # print("colors:", bic_colors)

        else:
            bic_colors = bicluster_colors
        bic_colors = dict(zip(biclusters.index.values, bic_colors))

        for row in biclusters.iterrows():
            bic_id = bic_prefix + str(row[0])
            s = row[1]["samples"]
            g = sorted(row[1]["genes_up"]) + sorted(row[1]["genes_down"])

            annot[bic_id] = "white"
            annot.loc[list(s), bic_id] = bic_colors[row[0]]

            if not no_row_colors:
                row_colors.loc[g, bic_id] = bic_colors[row[0]]

            g = [x for x in g if not x in ordered_genes]
            ordered_genes += g
            bic_names.append(bic_id)
    else:
        plot_bg_genes = True

    if plot_bg_genes:
        ordered_genes = ordered_genes + sorted(
            set(exprs.index.values).difference(set(ordered_genes))
        )

    col_colors = annot.loc[:, bic_names + cols]

    for col in reversed(cols):
        col_color_map = color_dict[col]
        col_colors[col] = col_colors[col].apply(lambda x: col_color_map[x])
        for subt in list(col_color_map.keys()):
            subt_samples = annot.loc[annot[col] == subt, :].index
            if cluster_columns:
                new_sample_order = [
                    x for x in sample_order if x not in subt_samples
                ] + [x for x in sample_order if x in subt_samples]
                sample_order = new_sample_order

    for col in reversed(bic_names):
        ordered_colors = [x for x in set(annot[col]) if not x == "white"] + ["white"]
        for subt in ordered_colors:
            subt_samples = annot.loc[annot[col] == subt, :].index
            if cluster_columns:
                new_sample_order = [
                    x for x in sample_order if x not in subt_samples
                ] + [x for x in sample_order if x in subt_samples]
                sample_order = new_sample_order

    # change column colors if redblue is chosen
    if bicluster_colors == "redblue":
        bic_ids = biclusters.loc[biclusters["direction"] == "UP", :].index.values
        bic_ids = [bic_prefix + str(bic_id) for bic_id in bic_ids]
        d = {"black": "red", "white": "blue"}
        col_colors.loc[:, bic_ids] = col_colors.loc[:, bic_ids].applymap(lambda x: d[x])
        bic_ids = biclusters.loc[biclusters["direction"] == "DOWN", :].index.values
        bic_ids = [bic_prefix + str(bic_id) for bic_id in bic_ids]
        d = {"black": "blue", "white": "red"}
        col_colors.loc[:, bic_ids] = col_colors.loc[:, bic_ids].applymap(lambda x: d[x])

    if no_bic_columns:
        col_colors = col_colors.loc[:, cols]

    vmin, vmax = color_range

    g = sns.clustermap(
        exprs.loc[ordered_genes, sample_order],
        figsize=figsize,
        col_cluster=cluster_cols,
        row_cluster=cluster_rows,
        dendrogram_ratio=dendrogram_ratio,
        colors_ratio=colors_ratio,
        cmap=sns.color_palette("coolwarm", as_cmap=True),
        vmin=vmin,
        vmax=vmax,
        xticklabels=col_labels,
        yticklabels=row_labels,
        col_colors=col_colors,
        row_colors=row_colors,
    )
    ax = g.ax_heatmap
    ax.set_ylabel("")
    ax.set_xlabel(xlabel)

    g.ax_row_dendrogram.set_visible(False)
    # g.cax.set_position([.10, .2, .03, .45])
    # from https://stackoverflow.com/questions/47350879/seaborn-clustermap-subplots-adjust-cancels-colour-bar-relocation
    dendro_box = g.ax_row_dendrogram.get_position()
    dendro_box.x0 = (dendro_box.x0 + 2 * dendro_box.x1) / 3
    g.cax.set_position(dendro_box)
    # Move the ticks to the left (https://stackoverflow.com/a/36939552/1878788)
    g.cax.yaxis.set_ticks_position("left")

    if no_cbar:
        g.ax_cbar.set_visible(False)

    # If there are row labels to highlight, format them in bold
    if highlight_row_labels:
        for row_tick in g.ax_heatmap.get_yticklabels():
            if row_tick.get_text() in highlight_row_labels:
                row_tick.set_weight("bold")
                for bic_id in biclusters.index.values:
                    if row_tick.get_text() in biclusters.loc[bic_id, "genes"]:
                        if not row_labels_black:
                            row_tick.set_color(bic_colors[bic_id])
    if legend_n_cols > 0:
        legends = []
        n_patches = 0
        for col in cols:
            patches = []
            col_color_map = color_dict[col]
            # add patches only for groups found in annotation
            plot_groups = [
                x for x in col_color_map.keys() if x in set(annot[col].values)
            ]
            for group in plot_groups:
                p = g.ax_col_dendrogram.bar(
                    0, 0, color=col_color_map[group], label=group, linewidth=0
                )
                patches.append(p)

            # add the legend
            n_patches += len(patches)
            l = g.ax_col_dendrogram.legend(
                loc="best", ncol=legend_n_cols, bbox_transform=gcf().transFigure
            )
            legends.append(l)
    return g, sample_order, (row_colors, col_colors)


def order_one(
    exprs, s0, subt_dict, subt_order=["Her2", "Basal", "LumA", "LumB", "Normal"]
):
    all_samples = set()
    for subt in subt_dict.keys():
        all_samples = all_samples.union(subt_dict[subt])

    s0 = set(s0)
    bg = all_samples.difference(s0)

    ordered_samples = []
    ordered_sets = [s0, bg]
    for s_i in ordered_sets:
        for subt in subt_order:
            ordered_subset = list(s_i.intersection(subt_dict[subt]))
            # ordered_subset = list(exprs.loc[:,ordered_subset].sum().sort_values().index)
            ordered_samples += ordered_subset
    return ordered_samples


def order_two(
    s0, s1, subt_dict, subt_order=["Her2", "Basal", "LumA", "LumB", "Normal"]
):
    all_samples = set()
    for subt in subt_dict.keys():
        all_samples = all_samples.union(subt_dict[subt])
    s0 = set(s0)
    s1 = set(s1)

    overlap = s0.intersection(s1)
    s0_ = s0.difference(s1)
    s1_ = s1.difference(s0)
    bg = all_samples.difference(s0).difference(s1)

    ordered_samples = []
    ordered_sets = [overlap, s0_, s1_, bg]
    for s_i in ordered_sets:
        for subt in subt_order:
            ordered_subset = list(s_i.intersection(subt_dict[subt]))
            ordered_samples += ordered_subset
    return ordered_samples



import numpy as np
import pandas as pd

from unpast.utils.similarity import get_similarity_jaccard

# from unpast.utils.similarity import get_similarity_corr


class TestGetSimilarityJaccard:
    """Test cases for get_similarity_jaccard function."""

    def test_jaccard_similarity_basic(self):
        """Test basic Jaccard similarity calculation."""
        # Create binary data with clear similarity patterns
        data = pd.DataFrame(
            {
                "gene1": [1, 1, 0, 0, 0],
                "gene2": [1, 1, 0, 0, 0],  # Identical to gene1
                "gene3": [0, 0, 1, 1, 1],  # Complement of gene1/gene2
                "gene4": [1, 0, 1, 0, 1],  # Different pattern
            },
            index=["s1", "s2", "s3", "s4", "s5"],
        )

        result = get_similarity_jaccard(data)

        # Check that it's symmetric
        assert result.shape == (4, 4)
        assert np.allclose(result.values, result.values.T)

        # Check diagonal is 1
        assert np.allclose(np.diag(result.values), 1.0)

        # gene1 and gene2 should be identical (similarity = 1)
        assert result.loc["gene1", "gene2"] == 1.0

        # Check that similarities are between 0 and 1
        assert np.all(result.values >= 0)
        assert np.all(result.values <= 1)

    def test_jaccard_similarity_complement_matching(self):
        """Test complement matching in Jaccard similarity."""
        # Create data where one gene has many 1s (should use complement)
        data = pd.DataFrame(
            {
                "gene1": [1, 1, 1, 1, 0],  # Many 1s
                "gene2": [0, 0, 0, 0, 1],  # Complement pattern
            },
            index=["s1", "s2", "s3", "s4", "s5"],
        )

        result = get_similarity_jaccard(data)

        # Should detect complement matching
        similarity_value = result.iloc[0, 1]  # Use iloc for numeric access
        assert similarity_value > 0

    def test_jaccard_similarity_single_gene(self):
        """Test with single gene."""
        data = pd.DataFrame(
            {
                "gene1": [1, 0, 1, 0, 1],
            },
            index=["s1", "s2", "s3", "s4", "s5"],
        )

        result = get_similarity_jaccard(data)

        assert result.shape == (1, 1)
        assert result.iloc[0, 0] == 1.0

    def test_jaccard_similarity_all_zeros(self):
        """Test with genes that have all zeros."""
        data = pd.DataFrame(
            {
                "gene1": [0, 0, 0, 0, 0],
                "gene2": [1, 0, 1, 0, 0],
            },
            index=["s1", "s2", "s3", "s4", "s5"],
        )

        result = get_similarity_jaccard(data)

        # Should handle edge cases gracefully
        assert result.shape == (2, 2)
        assert not np.any(np.isnan(result.values))


# class TestGetSimilarityCorr:
#     """Test cases for get_similarity_corr function."""

#     def test_correlation_similarity_basic(self):
#         """Test basic correlation similarity calculation."""
#         # Create data with known correlation patterns
#         data = pd.DataFrame(
#             {
#                 "gene1": [1.0, 2.0, 3.0, 4.0, 5.0],
#                 "gene2": [1.0, 2.0, 3.0, 4.0, 5.0],  # Perfect positive correlation
#                 "gene3": [5.0, 4.0, 3.0, 2.0, 1.0],  # Perfect negative correlation
#                 "gene4": [2.0, 1.0, 4.0, 3.0, 6.0],  # Different pattern
#             },
#             index=["s1", "s2", "s3", "s4", "s5"],
#         )

#         result = get_similarity_corr(data, verbose=False)

#         # Check basic properties
#         assert result.shape == (4, 4)
#         assert np.allclose(np.diag(result.values), 1.0)

#         # gene1 and gene2 should have perfect positive correlation
#         # Check that correlation matrix has the expected structure
#         assert result.shape == (4, 4)
#         assert result.iloc[0, 0] == 1.0  # Diagonal should be 1

#         # gene1 and gene3 should have negative correlation (set to 0)
#         assert result.loc["gene1", "gene3"] == 0.0

#         # All values should be >= 0 (negative correlations set to 0)
#         assert np.all(result.values >= 0)

#     def test_correlation_similarity_no_variation(self):
#         """Test with constant genes (no variation)."""
#         data = pd.DataFrame(
#             {
#                 "gene1": [1.0, 1.0, 1.0, 1.0, 1.0],  # No variation
#                 "gene2": [2.0, 3.0, 4.0, 5.0, 6.0],  # Normal variation
#             },
#             index=["s1", "s2", "s3", "s4", "s5"],
#         )

#         result = get_similarity_corr(data, verbose=True)

#         # Should handle constant genes gracefully
#         assert result.shape == (2, 2)
#         # Correlation with constant gene should be NaN, then filled with 0
#         assert result.loc["gene1", "gene2"] == 0.0



"""Binarization module for gene expression data."""

import sys
import warnings
from time import time

import numpy as np
import pandas as pd
from sklearn.cluster import AgglomerativeClustering, KMeans
from sklearn.mixture import GaussianMixture
from statsmodels.stats.multitest import fdrcorrection

from unpast import __version__
from unpast.utils.io import ProjectPaths, read_args, write_args
from unpast.utils.logs import get_logger, log_function_duration
from unpast.utils.statistics import calc_e_pval, calc_SNR, generate_null_dist, get_trend
from unpast.utils.visualization import plot_binarization_results, plot_binarized_feature

logger = get_logger(__name__)


def _fit_gmm_model(row, seed=42, prob_cutoff=0.5):
    """Fit a Gaussian Mixture Model to identify two groups in expression data.

    Args:
        row (array): expression values for a single feature across samples
        seed (int): random seed for reproducibility
        prob_cutoff (float): probability cutoff for group assignment

    Returns:
        tuple: (labels, is_converged) where
            - labels: boolean array indicating group membership
            - is_converged: whether the GMM fitting converged
    """
    with warnings.catch_warnings():  # this is to ignore convergence warnings
        warnings.simplefilter("ignore")
        row2d = row[:, np.newaxis]  # adding mock axis for GM interface
        model = GaussianMixture(
            n_components=2,
            init_params="kmeans",
            max_iter=len(row),
            n_init=1,
            covariance_type="spherical",
            random_state=seed,
        ).fit(row2d)

        labels = model.predict_proba(row2d)[:, 0] > prob_cutoff
        is_converged = model.converged_
        return labels, is_converged


def _select_pos_neg(row, min_n_samples, seed=42, prob_cutoff=0.5, method="GMM"):
    """Identify positive and negative signal groups using GMM or other method of binarization.

    Args:
        row (array): expression values for a single feature across samples
        min_n_samples (int): minimum number of samples required for each group
        seed (int): random seed for reproducible clustering
        prob_cutoff (float): probability threshold for group assignment
        method (str): binarization method to use ("GMM")

    Returns:
        tuple: (mask_pos, mask_neg, snr, size, is_converged) where
            - mask_pos: boolean mask for positive/high expression samples
            - mask_neg: boolean mask for negative/low expression samples
            - snr: signal-to-noise ratio between groups
            - size: effective sample size
            - is_converged: whether GMM fitting converged
    """
    is_converged = None
    if method == "GMM":
        labels, is_converged = _fit_gmm_model(row, seed=seed, prob_cutoff=prob_cutoff)

    elif method in ["kmeans", "ward"]:
        row2d = row[:, np.newaxis]  # adding mock axis
        if method == "kmeans":
            model = KMeans(n_clusters=2, max_iter=len(row), n_init=1, random_state=seed)
        else:
            assert method == "ward"
            model = AgglomerativeClustering(n_clusters=2, linkage="ward")
        # elif method == "HC_ward":
        #    model = Ward(n_clusters=2)

        labels = model.fit_predict(row2d) == 1

    else:
        raise NotImplementedError(
            f"wrong binarization method name {method},"
            " must be ['GMM', 'kmeans', 'ward']",
        )

    assert labels.dtype is np.dtype(bool)
    assert len(labels) == len(row)

    # let labels == True be always a smaller sample set
    if labels.sum() >= (~labels).sum():
        labels = ~labels

    # special treatment for cases when bic distribution is too wide and overlaps bg distribution
    # remove from bicluster samples with the sign different from its median sign
    # TODO: consider removing/using more robust method
    if labels.sum() > 0:
        if np.median(row[labels]) >= 0:
            labels[row < 0] = False
        else:
            labels[row > 0] = False

    if labels.sum() >= min_n_samples:
        size = labels.sum()  # the smaller group size
        snr = calc_SNR(row[labels], row[~labels])

        # return first the positive group
        if snr <= 0:
            labels = ~labels
        return labels, ~labels, abs(snr), size, is_converged

    else:
        # not enough samples in the group, return empty masks
        return (
            np.zeros_like(labels),
            np.zeros_like(labels),
            np.nan,
            np.nan,
            is_converged,
        )


def sklearn_binarization(
    exprs,
    min_n_samples,
    plot=True,
    plot_SNR_thr=2.0,
    show_fits=[],
    seed=1,
    prob_cutoff=0.5,
    method="GMM",
):
    """Perform binarization of gene expression data using GMM or other method.

    Args:
        exprs (DataFrame): expression matrix with genes as rows and samples as columns
        min_n_samples (int): minimum number of samples required for each group
        plot (bool): whether to generate plots for binarization
        plot_SNR_thr (float): SNR threshold above which to generate plots
        show_fits (list): specific gene names for which to show fitting plots
        seed (int): random seed for reproducible results
        prob_cutoff (float): probability threshold for group assignment
        method (str): binarization method to use ("GMM")

    Returns:
        tuple: (binarized_expressions, stats) where
            - binarized_expressions: dict mapping gene names to binary sample groups
            - stats: dict containing SNR and size statistics for each gene
    """
    binarized_expressions = {}
    stats = {}
    for i, (gene, row_) in enumerate(exprs.iterrows()):
        row = row_.values

        pos_mask, neg_mask, snr, size, is_converged = _select_pos_neg(
            row, min_n_samples, seed=seed, prob_cutoff=prob_cutoff, method=method
        )

        # bicluster is the smaller part
        if pos_mask.sum() <= neg_mask.sum():
            mask, direction = pos_mask, "UP"
        else:
            mask, direction = neg_mask, "DOWN"

        binarized_expressions[gene] = mask.astype(int)
        stats[gene] = {
            "pval": 0,
            "SNR": snr,
            "size": size,  # size of positive group
            "direction": direction,
            "convergence": is_converged,
        }

        # TODO: tqdm
        if i % 1000 == 0 and i > 0:
            logger.debug(f"processed {i}/{exprs.shape[0]} samples")

        if (plot and abs(snr) > plot_SNR_thr) or (gene in show_fits):
            plot_binarized_feature(row, direction, pos_mask, neg_mask, snr)

    stats = pd.DataFrame.from_dict(stats).T
    binarized_expressions = pd.DataFrame.from_dict(binarized_expressions)
    return binarized_expressions, stats


def _make_saveable_args(args: dict):
    """Prepare local variables for saving to file.
    Args:
        args (dict): Dictionary of arguments to prepare for saving
    Returns:
        dict: Prepared dictionary with only relevant keys for saving
    """
    args = args.copy()  # avoid modifying the original dictionary
    exprs = args["exprs"]
    args["exprs_hash"] = pd.util.hash_pandas_object(exprs).sum()
    del args["exprs"]  # remove large DataFrame from saving

    args["version"] = __version__  # save version of the code
    return args


def _check_same_args(paths: ProjectPaths, check_args: dict):
    """Check if the arguments match the saved binarization arguments.

    Args:
        paths (ProjectPaths): Object containing file paths for binarization files
        check_args (dict): Dictionary of arguments to match with saved files

    Returns:
        bool: True if arguments match, False otherwise
    """
    try:
        args = read_args(paths.binarization_args)
        args_str = {k: str(v) for k, v in args.items()}
        check_args_str = {k: str(v) for k, v in check_args.items()}

        # remove paths from comparison
        del args_str["paths"]
        del check_args_str["paths"]

        if args_str != check_args_str:
            diff = set(args_str.items()) ^ set(check_args_str.items())
            logger.warning(f"Cached binarization arguments differ: {diff}")
            return False

    except Exception:
        logger.exception(f"Failed to check cached arguments {paths.binarization_args}")
        return False

    return True


def _try_loading_binarization_files(paths: ProjectPaths, check_args: dict):
    """Try to load existing binarization files.

    Args:
        paths (ProjectPaths): Object containing file paths for binarization files
        check_args (dict): Dictionary of arguments to match with saved files

    Returns:
        tuple: (binarized_data, stats, null_distribution) where
            - binarized_data: DataFrame with binary expression profiles or None if failed
            - stats: DataFrame with binarization statistics or None if failed
            - null_distribution: DataFrame with background distribution or None if failed
    """
    if paths.bin_dir.exists():
        logger.info(
            f"Binarization directory {paths.bin_dir} already exist,"
            " trying to load precalculated binarization files."
        )
    else:
        logger.debug(
            f"Binarization directory {paths.bin_dir} does not exist yet,"
            " skipping loading precalculated binarization files."
        )
        return None, None, None

    # Check if the arguments match
    if not _check_same_args(paths, check_args):
        logger.warning(f"Ignoring cached binarization {paths}")
        return None, None, None

    binarized_data, stats, null_distribution = None, None, None
    logger.debug("Loading binarization files ...")

    # Try to load binarized data
    try:
        logger.debug(f"Loading binarized features from {paths.binarization_res}")
        binarized_data = pd.read_csv(paths.binarization_res, sep="\t", index_col=0)
    except Exception as e:
        logger.error(f"Failed to load {paths.binarization_res}, {e}")

    # Try to load stats
    try:
        logger.debug(f"Loading binarization statistics from {paths.binarization_stats}")
        stats = pd.read_csv(paths.binarization_stats, sep="\t", index_col=0)
    except Exception as e:
        logger.error(f"Failed to load {paths.binarization_stats}, {e}")

    # Try to load background distribution
    try:
        logger.debug(f"Loading background distribution from {paths.binarization_bg}")
        null_distribution = pd.read_csv(paths.binarization_bg, sep="\t", index_col=0)
        null_distribution.columns = [int(x) for x in null_distribution.columns.values]
    except Exception as e:
        logger.error(f"Failed to load {paths.binarization_bg}, {e}")
    return binarized_data, stats, null_distribution


def _save_binarization_files(
    paths, args_saveable, binarized_data, stats, null_distribution
):
    """Save binarization results to files.

    Args:
        paths (ProjectPaths): Object containing file paths for binarization files
        binarized_data (DataFrame): DataFrame with binary expression profiles
        stats (DataFrame): DataFrame with binarization statistics
        null_distribution (DataFrame): DataFrame with background distribution

    Returns:
        None: Saves files to specified paths
    """
    paths.create_binarization_paths()
    write_args(
        args_saveable,
        paths.binarization_args,
        args_label="Binarization arguments",
    )

    binarized_data.to_csv(paths.binarization_res, sep="\t")
    logger.debug(f"Binarized data saved to {paths.binarization_res}")

    stats.to_csv(paths.binarization_stats, sep="\t")
    logger.debug(f"Statistics saved to {paths.binarization_stats}")

    null_distribution = null_distribution.sort_index()
    null_distribution.to_csv(paths.binarization_bg, sep="\t")
    logger.debug(f"Background distribution saved to {paths.binarization_bg}")


def _generate_missing_null_distribution_sizes(
    null_distribution, N, sizes, pval, n_permutations, seed
):
    """Ensure that the null distribution contains all required sizes."""
    if null_distribution is None:
        precomputed_sizes = set()
    else:
        precomputed_sizes = set(null_distribution.index.values)

    add_sizes = np.array(sorted(set(sizes) - (set(precomputed_sizes))))
    if len(add_sizes) > 0:
        null_distribution2 = generate_null_dist(
            N,
            add_sizes,
            pval=pval,
            n_permutations=n_permutations,
            seed=seed,
        )
        null_distribution2.columns = [int(x) for x in null_distribution2.columns.values]

        if null_distribution is None:
            null_distribution = null_distribution2
        else:
            # join
            null_distribution = pd.concat(
                [
                    null_distribution,
                    null_distribution2.loc[:, null_distribution.columns.values],
                ],
                axis=0,
            )
    return null_distribution


def _add_snrs(stats, null_distribution, sizes, pval):
    """Calculate SNR p-values and thresholds based on null distribution.

    Args:
        stats (DataFrame): DataFrame with statistics including 'SNR' and 'size'
        null_distribution (DataFrame): DataFrame with empirical null distribution
    """
    stats = stats.dropna(subset=["size"])
    stats["pval"] = stats.apply(
        lambda row: calc_e_pval(row["SNR"], row["size"], null_distribution), axis=1
    )
    _accepted, pval_adj = fdrcorrection(stats["pval"])
    stats["pval_BH"] = pval_adj

    # find SNR threshold
    thresholds = np.quantile(null_distribution.loc[sizes, :].values, q=1 - pval, axis=1)
    size_snr_trend = get_trend(sizes, thresholds, plot=False)
    stats["SNR_threshold"] = stats["size"].apply(lambda x: size_snr_trend(x))
    return stats, size_snr_trend


@log_function_duration(name="Binarization")
def binarize(
    paths,
    exprs,
    method="GMM",
    min_n_samples=5,
    pval=0.001,
    plot_all=True,
    plot_SNR_thr=np.inf,
    show_fits=[],
    seed=42,
    prob_cutoff=0.5,
    n_permutations=10000,
    no_binary_save=False,
):
    """Main binarization function that creates binary expression profiles with significance testing.

    Args:
        out_dir (str | ProjectPaths): output directory or ProjectPaths object for saving results
        exprs (DataFrame): normalized expression matrix with features as rows and samples as columns
        method (str): binarization method to use ("GMM")
        save (bool): whether to save binarized data to files
        load (bool): whether to try loading existing binarized data
        min_n_samples (int): minimum number of samples required for each group
        pval (float): p-value threshold for significance testing
        plot_all (bool): whether to generate plots for all features
        plot_SNR_thr (float): SNR threshold above which to generate plots
        show_fits (list): specific feature names for which to show fitting plots
        seed (int): random seed for reproducible results
        prob_cutoff (float): probability threshold for group assignment
        n_permutations (int): number of permutations for null distribution generation

    Returns:
        tuple: (binarized_data, stats, null_distribution) where
            - binarized_data: DataFrame with binary expression profiles
            - stats: DataFrame with binarization statistics (SNR, size, direction)
            - null_distribution: DataFrame containing empirical null distribution for significance testing
    """
    args_saveable = _make_saveable_args(locals())
    binarized_data, stats, null_distribution = _try_loading_binarization_files(
        paths, check_args=args_saveable
    )

    # Calculate binarization data and statistics if not loaded
    if (binarized_data is None) or (stats is None):
        assert exprs is not None, (
            "No exprs provided and failed to load cached binarization data."
        )

        # binarize features
        binarized_data, stats = sklearn_binarization(
            exprs,
            min_n_samples,
            plot=plot_all,
            plot_SNR_thr=plot_SNR_thr,
            prob_cutoff=prob_cutoff,
            show_fits=show_fits,
            seed=seed,
            method=method,
        )

    # bicluster sizes for null distribution
    assert exprs is not None, "exprs must be defined"
    N = exprs.shape[1]

    def calc_size():
        assert stats is not None, "stats must be defined"
        sizes1 = set([x for x in stats["size"].values if not np.isnan(x)])
        # no more than 100 of bicluster sizes are computed
        # step = max(int((N - min_n_samples) / 100), 1)
        step = max(int((int(N / 2) - min_n_samples) / 100), 1)
        # sizes2 = set(map(int, np.arange(min_n_samples, int(N / 2), step)))
        sizes2 = set(map(int, np.arange(min_n_samples, int(N / 2) + 1, step)))
        sizes = np.array(sorted(sizes1 | sizes2))
        return sizes

    sizes = calc_size()

    # If null distribution was loaded, ensure all required sizes
    null_distribution = _generate_missing_null_distribution_sizes(
        null_distribution, N, sizes, pval, n_permutations, seed
    )

    # remove data for processing
    null_distribution_full = null_distribution
    null_distribution = null_distribution.loc[sizes, :]

    stats, size_snr_trend = _add_snrs(stats, null_distribution, sizes, pval)

    ### keep features passed binarization
    passed = stats.loc[stats["SNR"] > stats["SNR_threshold"], :]
    # passed = stats.loc[stats["pval_BH"]<pval,:]

    logger.debug(
        f"UP-regulated features:\t{passed.loc[passed['direction'] == 'UP', :].shape[0]}"
    )
    logger.debug(
        f"DOWN-regulated features:\t{passed.loc[passed['direction'] == 'DOWN', :].shape[0]}"
    )
    # print("ambiguous features:\t%s"%(passed.loc[passed["direction"]=="UP,DOWN",:].shape[0]),file = sys.stdout)

    # keep only binarized features
    binarized_data = binarized_data.loc[:, list(passed.index.values)]
    # add sample names
    binarized_data.index = exprs.columns.values

    if not no_binary_save:
        _save_binarization_files(
            paths,
            args_saveable,
            binarized_data,
            stats,
            null_distribution_full,  # save even not-used distributions
        )

    if plot_all:
        plot_binarization_results(stats, size_snr_trend, sizes, pval)

    return binarized_data, stats, null_distribution



import logging
import sys
import time
from functools import wraps
from typing import Union

LOG_LEVELS = {
    "CRITICAL": logging.CRITICAL,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "NOTSET": logging.NOTSET,
}

GLOBAL_PREFIX = ""  # for log_duration_decorator
INDENT = "  "


class PrefixLogger:
    """Wrapper around logging.Logger with prefix functionality"""

    def __init__(self, logger):
        self.logger = logger

    def info(self, msg, *args, **kwargs):
        """Log an info message with the global prefix"""
        self.logger.info(f"{GLOBAL_PREFIX} {msg}", *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        """Log a debug message with the global prefix"""
        self.logger.debug(f"{GLOBAL_PREFIX} {msg}", *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """Log a warning message with the global prefix"""
        self.logger.warning(f"{GLOBAL_PREFIX} {msg}", *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """Log an error message with the global prefix"""
        self.logger.error(f"{GLOBAL_PREFIX} {msg}", *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """Log a critical message with the global prefix"""
        self.logger.critical(f"{GLOBAL_PREFIX} {msg}", *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        """Log an exception message with the global prefix"""
        self.logger.exception(f"{GLOBAL_PREFIX} {msg}", *args, **kwargs)


def get_logger(name: Union[str, None] = None) -> PrefixLogger:
    """Get a logger with the specified name.

    Args:
        name (str, optional): Name of the logger.
    """
    return PrefixLogger(logging.getLogger(name))


def setup_logging(log_file=None, log_level=logging.INFO, log_file_level=logging.DEBUG):
    """Set up logging configuration.
    log_level messages will be logged to the console (stdout)
    log_file_level messages will be logged to the specified log file if provided.

    Args:
        log_file (str): Path to the log file. If None, no file logging is done.
        log_level: Logging level for console output.
        log_file_level: Logging level for file output.
    """
    # Create a custom logger
    logger = logging.getLogger()
    logger.setLevel(
        logging.DEBUG
    )  # Set to lowest level to allow all handlers to filter

    # Clear any existing handlers to avoid duplicates
    logger.handlers.clear()

    # Create formatter
    formatter = logging.Formatter(
        "%(asctime)s %(relativeCreated)8dms %(levelname)7s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%SUTC",
    )
    formatter.converter = time.gmtime  # Use UTC time for log timestamps

    # Create console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(log_level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    # Create file handler if log_file is specified
    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(log_file_level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    logger = logging.getLogger(__name__)
    return logger


def log_function_duration(name=None, indent=INDENT):
    """
    Decorator to log the duration of a function call.

    Usage:
        @log_function_duration(name="MyFunction")
        def my_function():
            # Function implementation
            pass

        @log_function_duration()
        def my_function():
            # Function implementation
            pass

    Args:
        name (str, optional): Name to use in the log message.
            If not provided, the function's name will be used.
        indent (str, optional): Indentation to apply to the log message.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Get logger for the module where the decorated function is defined
            func_logger = get_logger(func.__module__)
            if name:
                step_name = name
            else:
                step_name = func.__name__

            func_logger.debug(f"{step_name} started...")

            global GLOBAL_PREFIX
            GLOBAL_PREFIX = indent + GLOBAL_PREFIX

            start_time = time.time()
            result = func(*args, **kwargs)
            duration = time.time() - start_time

            GLOBAL_PREFIX = GLOBAL_PREFIX[len(indent) :]
            func_logger.debug(f"{step_name} completed in: {duration:.2f} seconds")
            return result

        return wrapper

    return decorator


@log_function_duration(name="Function Test")
def _test_func():
    """A simple test function to demonstrate logging."""
    time.sleep(0.5)  # Simulate some processing time
    return "Test function completed."


# initialize logging with default settings
setup_logging()
logger = get_logger(__name__)



import sys

import numpy as np
import pandas as pd

from unpast.utils.logs import get_logger

logger = get_logger(__name__)


def zscore(df):
    """Standardize expression data by z-score normalization.

    Args:
        df (DataFrame): input expression matrix with features as rows and samples as columns

    Returns:
        DataFrame: z-score normalized expression matrix
    """
    m = df.mean(axis=1)
    df = df.T - m
    df = df.T
    s = df.std(axis=1)
    df = df.T / s
    df = df.T
    # set to 0 not variable genes
    zero_var_genes = s[s == 0].index.values
    if len(zero_var_genes) > 0:
        logger.warning(
            f"{len(zero_var_genes)} zero variance rows detected, assign zero z-scores "
        )
    df.loc[zero_var_genes, :] = 0
    return df


def prepare_input_matrix(
    input_matrix: pd.DataFrame,
    min_n_samples: int = 5,
    tol: float = 0.01,
    standradize: bool = True,
    ceiling: float = 0,  # if float>0, limit z-scores to [-x,x]
):
    """Prepare and standardize input expression matrix for biclustering analysis.

    Args:
        input_matrix (DataFrame): raw expression matrix with features as rows and samples as columns
        min_n_samples (int): minimum number of samples required for processing
        tol (float): tolerance for checking if data is already standardized
        standradize (bool): whether to perform z-score standardization
        ceiling (float): if >0, limit z-scores to [-ceiling, ceiling] range

    Returns:
        DataFrame: processed and standardized expression matrix
    """
    exprs = input_matrix.copy()
    exprs.index = [str(x) for x in exprs.index.values]
    exprs.columns = [str(x) for x in exprs.columns.values]
    m = exprs.mean(axis=1)
    std = exprs.std(axis=1)
    # find zero variance rows
    zero_var = list(std[std == 0].index.values)
    if len(zero_var) > 0:
        logger.debug(f"Zero variance rows will be dropped: {len(zero_var)}")
        exprs = exprs.loc[std > 0]
        m = m[std > 0]
        std = std[std > 0]
        if exprs.shape[0] <= 2:
            logger.warning(
                "After excluding constant features (rows), less than 3 features (rows) remain"
                f" in the input matrix. Remaining: {exprs.shape[0]}"
            )

    mean_passed = np.all(np.abs(m) < tol)
    std_passed = np.all(np.abs(std - 1) < tol)
    if not (mean_passed and std_passed):
        logger.debug("Input is not standardized.")
        if standradize:
            exprs = zscore(exprs)
            if not mean_passed:
                logger.debug("- centering mean to 0")
            if not std_passed:
                logger.debug("- scaling std to 1")
    if len(set(exprs.index.values)) < exprs.shape[0]:
        logger.warning("Row names are not unique.")
    missing_values = exprs.isna().sum(axis=1)
    n_na = missing_values[missing_values > 0].shape[0]
    if n_na > 0:
        logger.warning(
            f"Missing values detected in {missing_values[missing_values > 0].shape[0]} rows"
        )
        keep_features = missing_values[
            missing_values <= exprs.shape[1] - min_n_samples
        ].index.values
        logger.warning(
            f"Features with too few values (<{min_n_samples}) dropped: {exprs.shape[0] - len(keep_features)}"
        )
        exprs = exprs.loc[keep_features, :]

    if standradize:
        if ceiling > 0:
            logger.debug(
                f"Standardized expressions will be limited to [-{ceiling},{ceiling}]:"
            )
            exprs[exprs > ceiling] = ceiling
            exprs[exprs < -ceiling] = -ceiling
            if n_na > 0:
                exprs.fillna(-ceiling, inplace=True)
                logger.debug(f"Missing values will be replaced with -{ceiling}.")
    return exprs



#!/usr/bin/env python3
import argparse

import numpy as np
import pandas as pd

from unpast import __version__
from unpast.core.binarization import binarize
from unpast.core.feature_clustering import run_Louvain, run_WGCNA, run_WGCNA_iterative
from unpast.core.preprocessing import prepare_input_matrix
from unpast.core.sample_clustering import make_biclusters
from unpast.utils.io import ProjectPaths, write_args, write_bic_table
from unpast.utils.logs import (
    LOG_LEVELS,
    get_logger,
    log_function_duration,
    setup_logging,
)
from unpast.utils.similarity import get_similarity_jaccard

logger = get_logger(__name__)


def check_input_shape(exprs_shape, min_n_samples):
    """Check if the input expression matrix has the required shape.

    Args:
        exprs_shape (tuple): shape of the input expression matrix (rows, columns)
        min_n_samples (int): minimum number of samples required

    Raises:
        AssertionError: if the input matrix does not meet the requirements
    """
    errs = []
    if exprs_shape[1] < 5:
        errs.append(
            "Input matrix must contain at least 5 samples (columns)"
            f" and 2 features (rows), but only {exprs_shape[1]}"
            " columns are found."
        )

    if exprs_shape[0] < 2:
        errs.append(
            "Input matrix must contain at least 2 features (rows)"
            f", but only {exprs_shape[0]} rows are found."
        )

    if min_n_samples < 2:
        errs.append(
            "The minimal number of samples in a bicluster `min_n_samples`"
            f" must be >= 2, found {min_n_samples}."
        )

    if min_n_samples > exprs_shape[1] // 2:
        errs.append(
            "The minimal number of samples in a bicluster `min_n_samples`"
            " must be not greater than half of the cohort size"
            f", found {min_n_samples} > {exprs_shape[1] // 2}."
        )

    if errs:
        for msg in errs:
            logger.error(msg)
        raise ValueError(errs[0])

    if min_n_samples < 5:
        logger.warning(
            f"min_n_samples is recommended to be >= 5, found {min_n_samples}"
        )


@log_function_duration(name="UnPaSt", indent="")
def unpast(
    exprs_file: str,
    out_dir: str = "runs/run_<timestamp>",
    binary_dir: str = "<out_dir>/binarization",
    no_binary_save: bool = False,
    ceiling: float = 3,
    bin_method: str = "kmeans",
    clust_method: str = "Louvain",
    min_n_samples: int = 5,
    show_fits: list = [],
    pval: float = 0.01,
    directions: list = ["DOWN", "UP"],
    modularity: float = 1 / 3,
    similarity_cutoffs=-1,  # for Louvain
    ds: int = 3,
    dch: float = 0.995,
    max_power: int = 10,
    precluster: bool = True,
    rpath: str = "",  # for WGCNA
    # cluster_binary: bool = False,
    merge: float = 1,
    seed: int = 42,
    verbose: bool = False,
    plot_all: bool = False,
    e_dist_size: int = 10000,
    standradize: bool = True,
):
    """Main UnPaSt biclustering algorithm for identifying differentially expressed biclusters.

    Args:
        exprs_file (str): expression matrix with features as rows and samples as columns
        out_dir (str): output directory path (default: "runs/run_<timestamp>")
        binary_dir (str): directory to save/load binarization results (default: "<out_dir>/binarization")
        no_binary_save (bool): if True, do not save binarization results (default: False)
        ceiling (float): absolute threshold for z-scores (default: 3)
        bin_method (str): binarization method - "kmeans", "ward", "GMM" (default: "kmeans")
        clust_method (str): clustering method - "WGCNA", "iWGCNA", "Louvain" (default: "Louvain")
        min_n_samples (int): minimal number of samples in bicluster (default: 5)
        show_fits (list): features to show binarization plots for (default: [])
        pval (float): binarization p-value threshold (default: 0.01)
        directions (list): clustering directions - ["DOWN","UP"] or ["BOTH"] (default: ["DOWN","UP"])
        modularity (float): modularity parameter for Louvain clustering (default: 1/3)
        similarity_cutoffs (float or list): similarity cutoffs for Louvain clustering (default: -1 for auto)
        ds (int): deepSplit parameter for WGCNA (default: 3)
        dch (float): detectCutHeight parameter for WGCNA (default: 0.995)
        max_power (int): maximum power for WGCNA (default: 10)
        precluster (bool): whether to precluster for WGCNA (default: True)
        rpath (str): path to Rscript executable for WGCNA (default: "")
        merge (float): similarity threshold for merging biclusters (default: 1)
        seed (int): random seed for reproducibility (default: 42)
        verbose (bool): whether to print progress messages (default: False)
        plot_all (bool): whether to show all plots (default: False)
        e_dist_size (int): size of empirical SNR distribution (default: 10000)
        standardize (bool): whether to standardize input matrix (default: True)

    Returns:
        pd.DataFrame: biclusters table with columns for genes, samples, SNR, etc.
    """
    np.random.seed(seed)  # todo: check if this is needed
    paths = ProjectPaths(out_dir=out_dir, binary_dir=binary_dir)

    # prepare logging and save args
    if verbose:
        setup_logging(log_file=paths.log, log_level=LOG_LEVELS["DEBUG"])
    else:
        setup_logging(log_file=paths.log, log_level=LOG_LEVELS["INFO"])
    version = __version__
    logger.debug(f"Running UnPaSt v. {version}")
    write_args(locals(), paths.args, args_label="Run arguments")

    # read inputs
    exprs = pd.read_csv(exprs_file, sep="\t", index_col=0)
    logger.info(f"Loaded input from {exprs_file}")
    logger.info(f"- input shape = {exprs.shape[0]} features x {exprs.shape[1]} samples")
    check_input_shape(exprs.shape, min_n_samples)

    e_dist_size = max(e_dist_size, int(1.0 / pval * 10))
    logger.debug(f"The size of empirical SNR distribution: {e_dist_size}")

    # check if input is standardized (between-sample)
    # if necessary, standardize and limit values to [-ceiling,ceiling]
    exprs = prepare_input_matrix(
        exprs,
        min_n_samples=min_n_samples,
        standradize=standradize,
        ceiling=ceiling,
    )

    ######### binarization #########

    binarized_features, stats, null_distribution = binarize(
        paths=paths,
        exprs=exprs,
        method=bin_method,
        no_binary_save=no_binary_save,
        min_n_samples=min_n_samples,
        pval=pval,
        plot_all=plot_all,
        show_fits=show_fits,
        seed=seed,
        prob_cutoff=0.5,
        n_permutations=e_dist_size,
    )

    bin_data_dict = {}
    stats = stats.loc[stats["pval"] <= pval, :]
    features_up = set(stats.loc[stats["direction"] == "UP", :].index.values)
    features_up = sorted(
        set(binarized_features.columns.values).intersection(set(features_up))
    )
    features_down = stats.loc[stats["direction"] == "DOWN", :].index.values
    features_down = sorted(
        set(binarized_features.columns.values).intersection(set(features_down))
    )

    df_up = binarized_features.loc[:, features_up]
    df_down = binarized_features.loc[:, features_down]
    if directions[0] == "BOTH":
        bin_data_dict["BOTH"] = pd.concat([df_up, df_down], axis=1)
    else:
        bin_data_dict["UP"] = df_up
        bin_data_dict["DOWN"] = df_down

    ######### gene clustering #########

    feature_clusters, not_clustered, used_similarity_cutoffs = [], [], []

    if clust_method == "Louvain":
        for d in directions:
            logger.debug(f"Clustering {d}-regulated features")
            df = bin_data_dict[d]
            if df.shape[0] > 1:
                similarity = get_similarity_jaccard(df)
                # similarity = get_similarity_corr(df,verbose = verbose)

                if similarity_cutoffs == -1:  # guess from the data
                    similarity_cutoffs = np.arange(0.3, 0.9, 0.01)
                # if similarity cuttofs is a single value turns it to a list
                try:
                    similarity_cutoffs = [elem for elem in similarity_cutoffs]
                except:
                    similarity_cutoffs = [similarity_cutoffs]

                # if modularity m is defined, choses a similarity cutoff corresponding to this modularity
                # and rund Louvain clustering
                modules, single_features, similarity_cutoff = run_Louvain(
                    similarity,
                    similarity_cutoffs=similarity_cutoffs,
                    m=modularity,
                )
                used_similarity_cutoffs.append(similarity_cutoff)
                feature_clusters += modules
                not_clustered += single_features
            elif df.shape[0] == 1:
                not_clustered += list(df.index.values)
                used_similarity_cutoffs.append(None)
        used_similarity_cutoffs = ",".join(map(str, used_similarity_cutoffs))

    elif "WGCNA" in clust_method:
        if clust_method == "iWGCNA":
            WGCNA_func = run_WGCNA_iterative
        else:
            WGCNA_func = run_WGCNA

        for d in directions:
            # WGCNA tmp file prefix
            logger.debug(f"Clustering {d}-regulated features")
            df = bin_data_dict[d]
            if df.shape[0] > 1:
                modules, single_features = WGCNA_func(
                    df,
                    paths=paths,
                    deepSplit=ds,
                    detectCutHeight=dch,
                    nt="signed_hybrid",
                    max_power=max_power,
                    precluster=precluster,
                    rpath=rpath,
                )
                feature_clusters += modules
                not_clustered += single_features

    else:
        raise NotImplementedError(
            f"'clust_method' must be 'WGCNA', 'iWGCNA', or 'Louvain', found {clust_method}"
        )

    ######### making biclusters #########
    if len(feature_clusters) == 0:
        logger.warning("No biclusters found")
        return pd.DataFrame()

    biclusters = make_biclusters(
        feature_clusters,
        binarized_features,
        exprs,
        method=bin_method,
        merge=merge,
        min_n_samples=min_n_samples,
        min_n_genes=2,
        seed=seed,
        cluster_binary=False,
    )

    ######### save biclusters #########
    if "WGCNA" in clust_method:
        modularity, similarity_cutoff = None, None
    elif clust_method == "Louvain":
        ds, dhs = None, None
    write_bic_table(
        biclusters,
        results_file_name=paths.res,
        to_str=True,
        add_metadata=True,
        seed=seed,
        min_n_samples=min_n_samples,
        pval=pval,
        bin_method=bin_method,
        clust_method=clust_method,
        directions=directions,
        similarity_cutoff=used_similarity_cutoffs,
        m=modularity,
        ds=ds,
        dch=dch,
        max_power=max_power,
        precluster=precluster,
        merge=merge,
    )

    # log results
    logger.info(f"Biclusters saved to {paths.res}")
    shapes = [
        (len(b["gene_indexes"]), len(b["sample_indexes"]))
        for _, b in biclusters.iterrows()
    ]
    shapes = sorted(shapes, key=lambda x: (x[0] * x[1]), reverse=True)
    log_str = ", ".join(f"{x}x{y}" for (x, y) in shapes[:10])
    if len(shapes) > 10:
        log_str += f"... and {len(shapes) - 10} more"
    logger.info(f"- found {len(biclusters)} biclusters with shapes: {log_str}")

    return biclusters


def parse_args():
    """Parse command line arguments for UnPaSt.

    Returns:
        Namespace: parsed command line arguments
    """
    parser = argparse.ArgumentParser(
        "UnPaSt identifies differentially expressed biclusters"
        f" in a 2-dimensional matrix. (version {__version__})"
    )
    parser.add_argument("--seed", metavar=42, default=42, type=int, help="random seed")
    parser.add_argument(
        "--exprs",
        metavar="exprs.z.tsv",
        required=True,
        help=".tsv file with between-sample normalized input data matrix. The first column and row must contain unique feature and sample ids, respectively. At least 5 samples (columns) and at least 2 features (rows) are required.",
    )
    parser.add_argument(
        "-o",
        "--out_dir",
        metavar="runs/run_<timestamp>",
        default="runs/run_<timestamp>",
        help="output folder",
    )
    parser.add_argument(
        "-bd",
        "--binary_dir",
        metavar="<out_dir>/binarization",
        default="<out_dir>/binarization",
        help="folder to save/load binarization results.",
    )
    parser.add_argument(
        "--ceiling",
        default=3,
        metavar="3",
        type=float,
        required=False,
        help="Absolute threshold for z-scores. For example, when set to 3, z-scores greater than 3 are set to 3 and z-scores less than -3 are set to -3. No ceiling if set to 0.",
    )
    parser.add_argument(
        "-s",
        "--min_n_samples",
        metavar=5,
        default=5,
        type=int,
        help="The minimal number of samples in a bicluster `min_n_samples` must be >= 2 and not greater than half of the cohort size.",
    )
    parser.add_argument(
        "-b",
        "--binarization",
        metavar="kmeans",
        default="kmeans",
        type=str,
        choices=["kmeans", "ward", "GMM"],
        help="binarization method",
    )
    parser.add_argument(
        "-p",
        "--pval",
        metavar=0.01,
        default=0.01,
        type=float,
        help="binarization p-value",
    )
    parser.add_argument(
        "-c",
        "--clustering",
        metavar="Louvain",
        default="Louvain",
        type=str,
        choices=["Louvain", "WGCNA"],
        help="feature clustering method",
    )
    # Louvain parameters
    parser.add_argument(
        "-m",
        "--modularity",
        default=1 / 3,
        metavar="1/3",
        type=float,
        help="Modularity corresponding to a cutoff for similarity matrix (Louvain clustering)",
    )
    parser.add_argument(
        "-r",
        "--similarity_cutoffs",
        default=-1,
        metavar="-1",
        type=float,
        help="A cutoff or a list of cuttofs for similarity matrix (Louvain clustering). If set to -1, will be chosen authomatically from [1/5,4/5] using elbow method.",
    )
    # WGCNA parameters
    parser.add_argument(
        "--ds",
        default=3,
        metavar="3",
        type=int,
        choices=[0, 1, 2, 3, 4],
        help="deepSplit parameter, see WGCNA documentation",
    )
    parser.add_argument(
        "--dch",
        default=0.995,
        metavar="0.995",
        type=float,
        help="dynamicTreeCut parameter, see WGCNA documentation",
    )
    parser.add_argument(
        "--bidirectional",
        action="store_true",
        help="Whether to cluster up- and down-regulated features together.",
    )
    parser.add_argument(
        "--rpath", default="", metavar="", type=str, help="Full path to Rscript."
    )
    # parser.add_argument('--merge', default=1, metavar="1", type=float,help = "Whether to merge biclustres similar in samples with Jaccard index not less than the specified.")
    parser.add_argument(
        "--no_binary_save",
        action="store_true",
        help="don't save binarization results (load only)",
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    # parser.add_argument('--plot', action='store_true', help = "show plots")

    return parser.parse_args()


def main():
    """Main entry point for UnPaSt command line interface.

    Parses command line arguments and runs the UnPaSt biclustering algorithm.
    """
    args = parse_args()

    directions = ["DOWN", "UP"]
    if args.bidirectional:
        directions = ["BOTH"]

    try:
        biclusters = unpast(
            args.exprs,
            out_dir=args.out_dir,
            binary_dir=args.binary_dir,
            no_binary_save=args.no_binary_save,
            ceiling=args.ceiling,
            bin_method=args.binarization,
            clust_method=args.clustering,
            pval=args.pval,
            directions=directions,
            min_n_samples=args.min_n_samples,
            show_fits=[],
            modularity=args.modularity,
            similarity_cutoffs=args.similarity_cutoffs,  # for Louvain
            ds=args.ds,
            dch=args.dch,
            rpath=args.rpath,
            precluster=True,  # for WGCNA
            # cluster_binary = False,
            # merge = args.merge,
            seed=args.seed,
            # plot_all = args.plot,
            verbose=args.verbose,
        )
    except Exception as e:
        logger.error(e)
        raise


if __name__ == "__main__":
    main()



"""Cluster binarized genes"""

import os
import subprocess
from pathlib import Path
from time import time

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from unpast.utils.logs import get_logger, log_function_duration

logger = get_logger(__name__)

RSCRIPTS_DIR = (Path(__file__).parent.parent / "rscripts").resolve()


@log_function_duration(name="WGCNA Iterative feature clustering")
def run_WGCNA_iterative(
    binarized_expressions,
    paths,
    deepSplit=0,
    detectCutHeight=0.995,
    nt="signed_hybrid",  # see WGCNA documentation
    max_power=10,
    precluster=False,
    rscr_path=False,
    rpath="",
):
    """Run WGCNA clustering iteratively until all features are clustered or stopping condition is met.

    Args:
        binarized_expressions (DataFrame): binary expression matrix with features as rows, samples as columns
        paths (ProjectPaths): project paths object containing paths for temporary files
        deepSplit (int): WGCNA parameter controlling module splitting sensitivity (0-4)
        detectCutHeight (float): WGCNA height cutoff for merging modules (0-1)
        nt (str): WGCNA network type ("signed_hybrid", "signed", "unsigned")
        max_power (int): maximum soft thresholding power to test in WGCNA
        precluster (bool): whether to perform pre-clustering before WGCNA
        rscr_path (bool): whether to use custom R script path
        rpath (str): path to R installation

    Returns:
        tuple: (modules, not_clustered) where
            - modules: list of feature modules/clusters found
            - not_clustered: list of features that could not be clustered
    """

    not_clustered = binarized_expressions.columns.values
    binarized_expressions_ = binarized_expressions.loc[:, :].copy()
    stop_condition = False

    modules = []
    i = 0
    while len(not_clustered) >= 3 and not stop_condition:
        binarized_expressions_ = binarized_expressions_.loc[:, not_clustered]

        m, not_clustered = run_WGCNA(
            binarized_expressions_,
            paths=paths,
            deepSplit=deepSplit,
            detectCutHeight=detectCutHeight,
            nt=nt,
            max_power=max_power,
            precluster=precluster,
            rscr_path=rscr_path,
            rpath=rpath,
        )
        logger.debug(
            f"WGCNA iteration {i}, modules:{len(m)}, not clustered:{len(not_clustered)}"
        )
        modules += m
        # stop when the number of not clustred samples does not change
        if len(m) == 0:
            stop_condition = True
            logger.debug(f"WGCNA iterations terminated at step {i}")

        i += 1
    return (modules, not_clustered)


@log_function_duration(name="WGCNA feature clustering")
def run_WGCNA(
    binarized_expressions,
    paths,
    deepSplit=0,
    detectCutHeight=0.995,
    nt="signed_hybrid",  # see WGCNA documentation
    max_power=10,
    precluster=False,
    rscr_path=False,
    rpath="",
):
    """Run WGCNA (Weighted Gene Co-expression Network Analysis) clustering on binarized expression data.

    Args:
        binarized_expressions (DataFrame): binary expression matrix with features as rows, samples as columns
        paths (ProjectPaths): project paths object containing paths for temporary files
        deepSplit (int): WGCNA parameter controlling module splitting sensitivity (0-4)
        detectCutHeight (float): WGCNA height cutoff for merging modules (0-1)
        nt (str): WGCNA network type ("signed_hybrid", "signed", "unsigned")
        max_power (int): maximum soft thresholding power to test in WGCNA
        precluster (bool): whether to perform pre-clustering before WGCNA
        rscr_path (bool): whether to use custom R script path
        rpath (str): path to R installation

    Returns:
        tuple: (modules, not_clustered) where
            - modules: list of feature modules/clusters found by WGCNA
            - not_clustered: list of features that could not be clustered
    """
    # create unique suffix for tmp files

    fname = paths.get_wgcna_tmp_file()

    logger.debug(f"WGCNA pre-clustering: {precluster}")
    if precluster:
        precluster = "T"
    else:
        precluster = "F"

    deepSplit = int(deepSplit)
    if not deepSplit in [0, 1, 2, 3, 4]:
        logger.error("deepSplit must be 1,2,3 or 4. See WGCNA documentation.")
        return ([], [])
    if not 0 < detectCutHeight < 1:
        logger.error(
            "detectCutHeight must be between 0 and 1. See WGCNA documentation."
        )
        return ([], [])
    logger.debug(f"Running WGCNA for {fname} ...")
    if not rscr_path:
        rscr_path = str(RSCRIPTS_DIR / "run_WGCNA.R")

    binarized_expressions_ = binarized_expressions.loc[:, :].copy()

    # add suffixes to duplicated feature names
    feature_names = binarized_expressions.columns.values
    duplicated_feature_ndxs = np.arange(binarized_expressions.shape[1])[
        binarized_expressions.columns.duplicated()
    ]

    if len(duplicated_feature_ndxs) > 0:
        new_feature_names = []
        for i in range(len(feature_names)):
            fn = feature_names[i]
            if i in duplicated_feature_ndxs:
                fn = str(fn) + "*" + str(i)
            new_feature_names.append(fn)
        logger.info(
            f"{len(duplicated_feature_ndxs)} duplicated feature names detected."
        )
        dup_fn_mapping = dict(zip(new_feature_names, feature_names))
        binarized_expressions_.columns = new_feature_names

    # replace spaces in feature names
    # otherwise won't parse R output
    feature_names = binarized_expressions.columns.values
    feature_names_with_space = [x for x in feature_names if " " in x]
    if len(feature_names_with_space) > 0:
        logger.debug(
            f"feature names containing spaces (will be replaced):{len(feature_names_with_space)}"
        )
        fn_mapping = {}
        fn_mapping_back = {}
        for fn in feature_names:
            if " " in fn:
                fn_ = fn.replace(" ", "_")
                fn_mapping[fn] = fn_
                fn_mapping_back[fn_] = fn
        binarized_expressions_ = binarized_expressions.rename(
            fn_mapping, axis="columns"
        )

    # save binarized expression to a file
    binarized_expressions_.to_csv(fname, sep="\t")

    # run Rscript
    if len(rpath) > 0:
        rpath = rpath + "/"

    r_cmd_args = [
        rpath + "Rscript",
        rscr_path,
        fname,
        str(deepSplit),
        str(detectCutHeight),
        nt,
        str(max_power),
        precluster,
    ]
    logger.debug("R command line: '" + " ".join(r_cmd_args) + "'")

    process = subprocess.Popen(
        r_cmd_args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.communicate()

    # log stdout and stderr
    if len(stdout) > 0:
        if len(stdout) > 100:
            stdout = str(stdout[:100] + b"...") + "(truncated)"
        else:
            stdout = str(stdout)
        logger.debug(f"WGCNA stdout: {stdout}")
    if len(stderr) > 0:
        logger.warning(f"WGCNA stderr: {stderr}")

    module_file = fname.replace(".tsv", ".modules.tsv")
    try:
        modules_df = pd.read_csv(module_file, sep="\t", index_col=0)
    except Exception as e:
        logger.warning(f"Failed to read output WGCNA file {module_file}, error: {e}")
        logger.warning(
            "- that may be ok on some inputs. But if you don't have R and WGCNA installed,"
            " please use the default -c Louvain method instead of -c WGCNA."
        )
        modules_df = pd.DataFrame.from_dict({})
        # TODO: avoid error in case of small error

    # read WGCNA output
    modules = []
    not_clustered = []
    module_dict = modules_df.T.to_dict()
    for i in module_dict.keys():
        genes = module_dict[i]["genes"].strip().split()
        # change feature names if they were modified

        # return spaces in feature names back if necessary
        if len(feature_names_with_space) > 0:
            for j in range(len(genes)):
                if genes[j] in fn_mapping_back.keys():
                    genes[j] = fn_mapping_back[genes[j]]
        # remove suffixes from duplicated feature names
        if len(duplicated_feature_ndxs) > 0:
            for j in range(len(genes)):
                if genes[j] in dup_fn_mapping.keys():
                    genes[j] = dup_fn_mapping[genes[j]]

        if i == 0:
            not_clustered = genes
        else:
            modules.append(genes)

    # remove WGCNA input and output files
    try:
        os.remove(module_file)
        paths.clear_wgcna_tmp_files()
    except Exception as e:
        logger.debug(f"Failed to remove WGCNA temporary files: {e}")

    logger.debug(
        f"Detected modules: {len(modules)}, not clustered features {len(not_clustered)} "
    )

    return (modules, not_clustered)


@log_function_duration(name="Louvain feature clustering")
def run_Louvain(
    similarity,
    similarity_cutoffs=np.arange(0.33, 0.95, 0.05),
    m=False,
    plot=False,
    modularity_measure="newman",
):
    """Run Louvain community detection clustering on similarity matrix.

    Args:
        similarity (DataFrame): feature similarity matrix
        similarity_cutoffs (array): range of similarity thresholds to test for clustering
        m (bool): whether to return additional modularity information
        plot (bool): whether to generate plots of modularity vs cutoffs
        modularity_measure (str): modularity measure to use ("newman", "dugue")

    Returns:
        tuple: (modules, not_clustered, best_Q) where
            - modules: list of feature modules/clusters found
            - not_clustered: list of features that could not be clustered
            - best_Q: best modularity score achieved
    """
    if similarity.shape[0] == 0:
        logger.error("no features to cluster")
        return [], [], None

    logger.debug("Running Louvain ...")
    logger.debug(f"modularity: {modularity_measure}")

    import sknetwork
    from sknetwork.clustering import Louvain

    try:
        from sknetwork.clustering import modularity

        old_sknetwork_version = True
    except:
        from sknetwork.clustering import get_modularity

        logger.debug(f"sknetwork version used: {sknetwork.__version__}")
        old_sknetwork_version = False
    try:
        from scipy.sparse.csr import csr_matrix
    except:
        from scipy.sparse import csr_matrix

    modularities = []
    feature_clusters = {}
    best_Q = np.nan
    for cutoff in similarity_cutoffs:
        # scan the whole range of similarity cutoffs
        # e.g. [1/4;9/10] with step 0.5
        sim_binary = similarity.copy()
        sim_binary[sim_binary < cutoff] = 0
        sim_binary[sim_binary != 0] = 1
        rsums = sim_binary.sum()
        non_zero_features = rsums[rsums > 0].index
        sim_binary = sim_binary.loc[non_zero_features, non_zero_features]
        gene_names = sim_binary.index.values
        sparse_matrix = csr_matrix(sim_binary)

        if old_sknetwork_version:
            labels = Louvain(modularity=modularity_measure).fit_transform(sparse_matrix)
            Q = modularity(sparse_matrix, labels)
        else:
            sparse_matrix = sparse_matrix.astype("bool")
            labels = Louvain(modularity=modularity_measure).fit_predict(sparse_matrix)
            Q = get_modularity(sparse_matrix, labels)

        modularities.append(Q)
        # if binary similarity matrix contains no zeroes
        # bugfix for Louvain()
        if sim_binary.min().min() == 1:
            labels = np.zeros(len(labels))
        feature_clusters[cutoff] = labels

    # if similarity_cutoffs contains only one value, choose it as best_cutoff
    if len(similarity_cutoffs) == 1:
        best_cutoff = similarity_cutoffs[0]
        best_Q = Q

    # find best_cutoff automatically
    else:
        # check if modularity(cutoff)=const
        if len(set(modularities)) == 1:
            best_cutoff = similarity_cutoffs[-1]
            best_Q = modularities[-1]
            labels = feature_clusters[best_cutoff]

        #  if modularity!= const, scan the whole range of similarity cutoffs
        #  e.g. [1/4;9/10] with step 0.05
        else:
            # find the knee point in the dependency modularity(similarity curoff)
            from kneed import KneeLocator

            # define the type of the curve
            curve_type = "increasing"
            if modularities[0] >= modularities[-1]:
                curve_type = "decreasing"
            logger.debug(f"curve type: {curve_type}")
            # detect knee and choose the one with the highest modularity
            try:
                kn = KneeLocator(
                    similarity_cutoffs,
                    modularities,
                    curve="concave",
                    direction=curve_type,
                    online=True,
                )
                best_cutoff = kn.knee
                best_Q = kn.knee_y
                labels = feature_clusters[best_cutoff]
            except:
                logger.error("Failed to identify similarity cutoff")
                logger.info(f"Similarity cutoff: set to {similarity_cutoffs[0]}")
                best_cutoff = similarity_cutoffs[0]
                best_Q = np.nan
                logger.info(f"Modularity: {modularities}")
                if plot:
                    plt.plot(similarity_cutoffs, modularities, "bx-")
                    plt.xlabel("similarity cutoff")
                    plt.ylabel("modularity")
                    plt.show()
                    # return [], [], None
            if m:
                # if upper threshold for modularity m is specified
                # chose the lowest similarity cutoff at which modularity reaches >= m
                best_cutoff_m = 1
                for i in range(len(modularities)):
                    if modularities[i] >= m:
                        best_cutoff_m = similarity_cutoffs[i]
                        best_Q_m = modularities[i]
                        labels_m = feature_clusters[best_cutoff]
                        break
                if best_cutoff_m < best_cutoff:
                    best_cutoff = best_cutoff_m
                    best_Q = best_Q_m
                    labels = labels_m

    if plot and len(similarity_cutoffs) > 1:
        plt.plot(similarity_cutoffs, modularities, "bx-")
        plt.vlines(
            best_cutoff, plt.ylim()[0], plt.ylim()[1], linestyles="dashed", color="red"
        )
        plt.xlabel("similarity cutoff")
        plt.ylabel("modularity")
        plt.show()

    modules = []
    not_clustered = []

    for label in set(labels):
        ndx = np.argwhere(labels == label).reshape(1, -1)[0]
        genes = gene_names[ndx]
        if len(genes) > 1:
            modules.append(genes)
        else:
            not_clustered.append(genes[0])
    logger.debug(
        f"Detected modules: {len(modules)}, not clustered features {len(not_clustered)} "
    )
    logger.debug(f"- similarity cutoff: {best_cutoff:.2f}")
    logger.debug(f"- modularity: {best_Q:.3f}")
    return modules, not_clustered, best_cutoff



"""Tests for eval module."""

import os
import tempfile

import pandas as pd
import pytest

from unpast.misc.eval import calculate_perfromance, generate_exprs
from unpast.utils.io import read_bic_table


class TestGenerateExprs:
    """Test cases for generate_exprs function."""

    def test_generate_exprs_biclusters_write_read_roundtrip(self):
        """Test that created true biclusters can be written and read correctly."""

        with tempfile.TemporaryDirectory() as temp_dir:
            _, biclusters_original, _ = generate_exprs(
                data_sizes=(10, 5),
                frac_samples=[0.2, 0.3],
                outdir=temp_dir + "/",
                outfile_basename="test",
                seed=42,
            )

            bic_file_path = os.path.join(temp_dir, "test.true_biclusters.tsv.gz")
            biclusters_read = read_bic_table(bic_file_path)

            pd.testing.assert_frame_equal(
                biclusters_original,
                biclusters_read,
                check_names=False,  # index.name are diffferent
                check_dtype=False,  # float vs np.float64
            )

    @pytest.mark.slow
    def test_generate_real_smoke(self, tmp_path):
        """Smoke test to verify generate_exprs runs without errors"""

        # Test parameters
        n_biomarkers = 500
        frac_samples = [0.05, 0.1, 0.25, 0.5]
        # dimensions of the matrix
        n_genes = 10000  # gemes
        N = 200  # samples
        m = 4
        std = 1
        seed = 42

        # Scenario configuration
        sc_name = "C"
        params = {
            "C": {
                "add_coexpressed": [500] * 4,
                "g_overlap": False,
                "s_overlap": True,
            }
        }

        scenario = f"{sc_name}_{n_biomarkers}"

        # Run the function
        data, ground_truth, coexpressed_modules = generate_exprs(
            (n_genes, N),
            g_size=n_biomarkers,
            frac_samples=frac_samples,
            m=m,
            std=std,  # int, not float?
            outdir=str(tmp_path) + "/",
            outfile_basename=scenario,
            g_overlap=params[sc_name]["g_overlap"],
            s_overlap=params[sc_name]["s_overlap"],
            seed=seed,
            add_coexpressed=params[sc_name]["add_coexpressed"],
        )

        # Basic assertions to verify function completed
        assert data is not None
        assert ground_truth is not None
        assert coexpressed_modules is not None

        # Verify output files were created
        expected_files = [
            f"{scenario}*data.tsv.gz",
            f"{scenario}*true_biclusters.tsv.gz",
        ]

        for template in expected_files:
            fs = list(tmp_path.glob(template))
            assert len(fs) == 1, "Unexpected amount of generated files"


class TestCalculatePerformance:
    """Test cases for calculate_perfromance function."""

    def test_calculate_performance_smoke_test(self, tmp_path):
        """Simple smoke test for calculate_perfromance function."""
        # 1) build same data
        n_biomarkers = 50
        frac_samples = [0.1, 0.25, 0.5]
        # dimensions of the matrix
        n_genes = 200  # gemes
        N = 20  # samples
        m = 4
        std = 1
        seed = 42

        # Scenario configuration
        sc_name = "C"
        params = {
            "C": {
                "add_coexpressed": [50] * 4,
                "g_overlap": False,
                "s_overlap": True,
            }
        }

        scenario = f"{sc_name}_{n_biomarkers}"

        # Run the function
        data, ground_truth, coexpressed_modules = generate_exprs(
            (n_genes, N),
            g_size=n_biomarkers,
            frac_samples=frac_samples,
            m=m,
            std=std,  # int, not float?
            outdir=str(tmp_path) + "/",
            outfile_basename=scenario,
            g_overlap=params[sc_name]["g_overlap"],
            s_overlap=params[sc_name]["s_overlap"],
            seed=seed,
            add_coexpressed=params[sc_name]["add_coexpressed"],
        )

        # Basic assertions to verify function completed
        assert data is not None
        assert ground_truth is not None
        assert coexpressed_modules is not None

        true_biclusters = list(tmp_path.glob("*true_biclusters*"))
        assert len(true_biclusters) == 1
        gt = read_bic_table(true_biclusters[0])

        # Use the ground truth biclusters as our sample_clusters for testing
        sample_clusters = gt.copy()

        # check gt versus gt (should give perfect performance):
        all_samples = set(data.columns)

        # Create known_groups based on the actual ground truth biclusters
        gt_known_groups = {"ground_truth": {}}
        for idx, row in gt.iterrows():
            gt_known_groups["ground_truth"][f"bic_{idx}"] = row["samples"]

        # Test ground truth vs ground truth - should give perfect performance
        gt_performances, gt_best_matches = calculate_perfromance(
            sample_clusters_=sample_clusters,
            known_groups=gt_known_groups,
            all_samples=all_samples,
        )

        # Assert perfect performance when comparing ground truth to itself
        assert isinstance(gt_performances, pd.Series), (
            "gt_performances should be a Series"
        )
        assert "ground_truth" in gt_performances.index, (
            "Should have ground_truth performance"
        )
        gt_score = gt_performances.iloc[0]  # Get first (and should be only) value
        # Convert to float if needed and check if it's close to 1.0
        if isinstance(gt_score, (int, float)):
            assert abs(float(gt_score) - 1.0) < 1e-10, (
                f"GT vs GT should give perfect score (1.0), got {gt_score}"
            )
        else:
            # Just verify we got some result (might be different type than expected)
            assert gt_score is not None, (
                f"GT vs GT should give a non-null result, got {gt_score}"
            )

        # check some arbitrary diff
        all_samples = set(data.columns)
        sample_list = list(all_samples)
        mid_point = len(sample_list) // 2

        known_groups = {
            "test_classification": {
                "group1": set(sample_list[:mid_point]),
                "group2": set(sample_list[mid_point:]),
            }
        }

        # Call the function - should not crash
        performances, best_matches = calculate_perfromance(
            sample_clusters_=sample_clusters,
            known_groups=known_groups,
            all_samples=all_samples,
        )

        # Basic checks that it returns expected types
        assert isinstance(performances, pd.Series), "performances should be a Series"
        assert isinstance(best_matches, pd.DataFrame), (
            "best_matches should be a DataFrame"
        )

        # Verify that we got results for our test classification
        assert "test_classification" in performances.index, (
            "Should have performance for test_classification"
        )

        # Verify best_matches has expected columns
        expected_columns = ["classification", "Jaccard", "weight"]
        for col in expected_columns:
            if col in best_matches.columns:
                assert col in best_matches.columns, (
                    f"best_matches should have {col} column"
                )



######## Make biclusters #########
import numpy as np
import pandas as pd
from sklearn.cluster import AgglomerativeClustering, KMeans
from sklearn.mixture import GaussianMixture

from unpast.core.feature_clustering import run_Louvain
from unpast.utils.logs import get_logger, log_function_duration
from unpast.utils.similarity import get_similarity_jaccard
from unpast.utils.statistics import calc_SNR

logger = get_logger(__name__)


def cluster_samples(data, min_n_samples=5, seed=0, method="kmeans"):
    """Cluster samples into bicluster and background groups using unsupervised methods.

    Args:
        data (DataFrame): expression data with samples as rows and features as columns
        min_n_samples (int): minimum number of samples required for each cluster
        seed (int): random seed for reproducible clustering
        method (str): clustering method to use ("kmeans", "Jenks", "ward")

    Returns:
        dict: bicluster information containing sample indices and cluster size
    """
    # identify identify bicluster and backgound groups using 2-means
    max_n_iter = max(max(data.shape), 500)
    if method == "kmeans" or method == "Jenks":
        labels = (
            KMeans(
                n_clusters=2,
                random_state=seed,
                init="random",
                n_init=10,
                max_iter=max_n_iter,
            )
            .fit(data)
            .labels_
        )
    elif method == "ward":
        labels = AgglomerativeClustering(n_clusters=2, linkage="ward").fit(data).labels_
    # elif method == "HC_ward":
    #        model = Ward(n_clusters=2).fit(data).labels_
    elif method == "GMM":
        labels = GaussianMixture(
            n_components=2,
            init_params="kmeans",
            max_iter=max_n_iter,
            n_init=5,
            covariance_type="spherical",
            random_state=seed,
        ).fit_predict(data)
    ndx0 = np.where(labels == 0)[0]
    ndx1 = np.where(labels == 1)[0]
    if min(len(ndx1), len(ndx0)) < min_n_samples:
        return {}
    if len(ndx1) > len(ndx0):
        samples = ndx0
    else:
        samples = ndx1

    bicluster = {"sample_indexes": set(samples), "n_samples": len(samples)}
    return bicluster


@log_function_duration(name="Building biclusters from modules")
def modules2biclusters(
    modules,
    data_to_cluster,
    method="kmeans",
    min_n_samples=5,
    min_n_genes=2,
    seed=0,
):
    """Convert feature modules to biclusters by identifying optimal sample sets for each module.

    Args:
        modules (list): list of feature modules/clusters from clustering algorithms
        data_to_cluster (DataFrame): expression data with samples as rows and features as columns
        method (str): sample clustering method to use ("kmeans", "Jenks", "ward", "GMM")
        min_n_samples (int): minimum number of samples required for each bicluster
        min_n_genes (int): minimum number of genes required for each bicluster
        seed (int): random seed for reproducible clustering

    Returns:
        dict: biclusters dictionary with bicluster IDs as keys and bicluster information as values
    """
    biclusters = {}
    i = 0

    for mid in range(0, len(modules)):
        genes = modules[mid]
        if len(genes) >= min_n_genes:
            # cluster samples in a space of selected genes
            data = data_to_cluster.loc[genes, :].T
            bicluster = cluster_samples(
                data, min_n_samples=min_n_samples, seed=seed, method=method
            )
            if len(bicluster) > 0:
                bicluster["id"] = i
                bicluster["genes"] = set(genes)
                bicluster["n_genes"] = len(bicluster["genes"])
                biclusters[i] = bicluster
                i += 1

    logger.debug(f"Identified optimal sample sets for {len(modules)} modules.")
    logger.debug(
        f"Passed biclusters (>={min_n_genes} genes, >= {min_n_samples} samples): {i - 1}"
    )

    return biclusters


def update_bicluster_data(bicluster, data):
    """Update bicluster with additional metadata including up/down-regulated genes and z-scores.

    Args:
        bicluster (dict): bicluster dictionary containing "genes" and "samples" or "sample_indexes"
        data (DataFrame): complete expression data with all features (not just binarized)

    Returns:
        dict: updated bicluster with additional fields:
            - "samples": sample names
            - "gene_indexes": gene indices
            - "genes_up": up-regulated genes
            - "genes_down": down-regulated genes
            - "SNR": SNR for absolute average z-scores of all bicluster genes
    """
    sample_names = data.columns.values
    gene_names = data.index.values

    # ensure "sample_indexes" is present
    if "sample_indexes" not in bicluster.keys():
        assert "samples" in bicluster.keys(), (
            '"samples" or "sample_indexes" of a bicluster not specified'
        )
        sample_mask = np.isin(sample_names, list(bicluster["samples"]))
        bicluster["sample_indexes"] = set(np.where(sample_mask)[0])

    # ensure "samples" is present
    if "samples" not in bicluster.keys():
        inds = list(bicluster["sample_indexes"])
        bicluster["samples"] = set(sample_names[inds])

    sample_mask = np.isin(
        np.arange(len(sample_names)), list(bicluster["sample_indexes"])
    )
    bic_samples = sample_names[sample_mask]
    bg_samples = sample_names[~sample_mask]

    gene_mask = np.isin(gene_names, list(bicluster["genes"]))
    bic_genes = gene_names[gene_mask]
    bicluster["gene_indexes"] = set(np.where(gene_mask)[0])

    # distinguish up- and down-regulated features
    m_bic = data.loc[bic_genes, bic_samples].mean(axis=1)
    m_bg = data.loc[bic_genes, bg_samples].mean(axis=1)
    genes_up = m_bic[m_bic >= m_bg].index.values
    genes_down = m_bic[m_bic < m_bg].index.values
    bicluster["genes_up"] = set(genes_up)
    bicluster["genes_down"] = set(genes_down)

    genes_up = m_bic[m_bic >= m_bg].index.values
    genes_down = m_bic[m_bic < m_bg].index.values

    # calculate average z-score for each sample
    if min(len(genes_up), len(genes_down)) > 0:  # take sign into account
        avg_zscore = (
            data.loc[genes_up, :].sum() - data.loc[genes_down, :].sum()
        ) / bicluster["n_genes"]
    else:
        avg_zscore = data.loc[sorted(bicluster["genes"]), :].mean()

    # compute SNR for average z-score for this bicluster
    bicluster["SNR"] = np.abs(
        calc_SNR(avg_zscore[bic_samples], avg_zscore[bg_samples], pd_mode=True)
    )
    return bicluster


@log_function_duration(name="Bicluster merging")
def merge_biclusters(
    biclusters, data, J=0.8, min_n_samples=5, seed=42, method="kmeans"
):
    """Merge biclusters with similar sample sets based on Jaccard similarity.

    Args:
        biclusters (dict): dictionary of biclusters to be merged
        data (DataFrame): expression data with samples as columns
        J (float): Jaccard similarity threshold for merging biclusters
        min_n_samples (int): minimum number of samples required for merged biclusters
        seed (int): random seed for reproducible clustering
        method (str): clustering method for sample grouping

    Returns:
        dict: merged biclusters dictionary with reduced redundancy
    """
    #  bicluaters -> binary -> jaccard sim
    binary_representation = {}
    N = data.shape[1]
    for i in biclusters.keys():
        b = np.zeros(N)
        s_ndx = list(biclusters[i]["sample_indexes"])
        b[s_ndx] = 1
        binary_representation[i] = b
    binary_representation = pd.DataFrame.from_dict(binary_representation)
    binary_representation.index = data.columns.values
    bic_similarity = get_similarity_jaccard(binary_representation)
    # bic_similarity[bic_similarity >= J] = 1
    # bic_similarity[bic_similarity < J] = 0
    # find groups of biclusters including the same sample sets
    merged, not_merged, similarity_cutoff = run_Louvain(
        bic_similarity, plot=False, similarity_cutoffs=[J]
    )
    if len(merged) == 0:
        logger.debug("No biclusters to merge")
        return biclusters

    merged_biclusters = {}
    # add biclusters with no changes
    for bic_id in not_merged:
        merged_biclusters[bic_id] = biclusters[bic_id]

    # merge biclusters with overlapping sample sets
    for bic_group in merged:
        bic_group = sorted(bic_group)
        logger.debug(f"merged biclusters {bic_group}")
        new_bicluster = biclusters[bic_group[0]]
        # update genes
        for bic_id in bic_group[1:]:
            bic2 = biclusters[bic_id]
            new_bicluster["genes"] = new_bicluster["genes"] | bic2["genes"]
            new_bicluster["n_genes"] = len(new_bicluster["genes"])
        # update sample set for new bicluster
        # cluster samples in a space of selected genes
        new_bicluster.update(
            cluster_samples(
                data.loc[list(new_bicluster["genes"]), :].T,
                min_n_samples=min_n_samples,
                seed=seed,
                method=method,
            )
        )
        new_bicluster["n_samples"] = len(new_bicluster["sample_indexes"])
        merged_biclusters[bic_group[0]] = new_bicluster
    return merged_biclusters


@log_function_duration(name="Creating final biclusters")
def make_biclusters(
    feature_clusters,
    binarized_data,
    data,
    merge=1,
    min_n_samples=10,
    min_n_genes=2,
    method="kmeans",
    seed=42,
    cluster_binary=False,
):
    """Create final biclusters from feature clusters by adding sample information and performing merging.

    Args:
        feature_clusters (list): list of feature modules/clusters from clustering algorithms
        binarized_data (DataFrame): binary expression matrix
        data (DataFrame): original z-score normalized expression data
        merge (float): Jaccard similarity threshold for merging biclusters (0-1)
        min_n_samples (int): minimum number of samples required for each bicluster
        min_n_genes (int): minimum number of genes required for each bicluster
        method (str): sample clustering method ("kmeans", "ward", "GMM")
        seed (int): random seed for reproducible clustering
        cluster_binary (bool): whether to cluster on binary data (True) or z-scores (False)

    Returns:
        DataFrame: final biclusters with metadata including genes, samples, SNR, and z-scores
    """

    biclusters = []

    if cluster_binary:
        data_to_cluster = binarized_data.loc[:, :].T  # binarized expressions
    else:
        data_to_cluster = data.loc[binarized_data.columns.values, :]  # z-scores

    if len(feature_clusters) == 0:
        logger.warning("No biclusters found.")
    else:
        biclusters = modules2biclusters(
            feature_clusters,
            data_to_cluster,
            method=method,
            min_n_samples=min_n_samples,
            min_n_genes=min_n_genes,
            seed=seed,
        )

        ### merge biclusters with highly similar sample sets
        if merge < 1.0:
            biclusters = merge_biclusters(
                biclusters,
                data,
                method=method,
                J=merge,
                min_n_samples=min_n_samples,
                seed=seed,
            )

        for i in list(biclusters.keys()):
            biclusters[i] = update_bicluster_data(biclusters[i], data)

    biclusters = pd.DataFrame.from_dict(biclusters).T
    # add direction
    biclusters["direction"] = "BOTH"
    biclusters.loc[
        biclusters["n_genes"] == biclusters["genes_up"].apply(len), "direction"
    ] = "UP"
    biclusters.loc[
        biclusters["n_genes"] == biclusters["genes_down"].apply(len), "direction"
    ] = "DOWN"

    # add p-value for bicluster SNR (computed for avg. zscores)
    # use the same distribution as for single features
    # biclusters["e_pval"] = biclusters.apply(lambda row: calc_e_pval(row["SNR"], row["n_samples"], null_distribution),axis=1)

    # sort and reindex
    biclusters = biclusters.sort_values(by=["SNR", "n_genes"], ascending=[False, False])
    biclusters.index = range(0, biclusters.shape[0])

    biclusters = biclusters.loc[
        :,
        [
            "SNR",
            "n_genes",
            "n_samples",
            "genes",
            "samples",
            "direction",
            "genes_up",
            "genes_down",
            "gene_indexes",
            "sample_indexes",
        ],
    ]

    return biclusters
